###############################################################################
# File: constants_generator.py
# Author: Cameron Smith
# Date of Last Edit: June 2, 2017
#
# Purpose: Generate the necessary constants used in pvcam.h into a reusable
#          python module.
#
# Notes: No formal testing exists for this script. Please report all bugs to
#        to whoever is in charge of this project currently.
#
# Bugs: Enums are not created as Enum objects. Any struct that assigns to an
#       enum will result in a NameError as the Enum does not exist; only its
#       entities. Update occurrences manually to be a ctypes C type until fix.
#       (Known occurrence(s): md_ext_item_info [struct]. Currently patched by
#       assigning values to be of type `void`.
###############################################################################
import os
import re
from datetime import date

pvcam_sdk_path = os.environ["PVCAM_SDK_PATH"]
pvcam = r'{}inc/pvcam.h'.format(pvcam_sdk_path)
constants = r'./pyvcam/constants.py'


def define_writer(match):
    """Constructs a string representing defines in header files.

    Parameters:
        match: The matched pattern from the regular expression containing the
               named groups of the variable name and its associated value.
    """

    return match.group('var'), match.group('val')


def enum_writer(match, infile):
    """Constructs a tuple containing the data to represent enums in python.

    The tuple will have the form (<String>, <List>), where the string will
    contain the comment block denoting which enum the following list is from,
    and the list contains tuples pairing variables to values.

    Paramters:
        match: The matched pattern from the regular expression containing the
               name of the current enum.
        infile: The file being read from.
    """

    enums = []
    position = 0  # The current position inside of `enums`
    enum_group = match.group('name')

    # Skip two lines ahead to ignore the '{' line
    next(infile)
    line = next(infile)
    line = remove_comment(line, infile)

    # Set up the regular expressions to find name and values of enums
    enum_name = '\s+(?P<name>\w+)'
    enum_val = enum_name + '\s*=(?P<val>[^,]*)'

    # Loop through the file until the end of the current enum is reached.
    # If the current line matches the pattern <name> = <val>, then append
    # a tuple onto `enums` that pairs <name> with <val>. If the current
    # line does not match that pattern, check if it matches the pattern
    # <name>. If it does, append a tuple to `enums` pairing <name> with
    # the previous position's <name> + 1. If position = 0, then it is the
    # first element in the enum, so pair <name> with 0 instead. If the
    # line does not match either of these patters, move to the next line.
    while line != '}\n':
        val = re.search(enum_val, line)
        if val:
            enums.append((val.group('name'), val.group('val')))
        else:
            name = re.search(enum_name, line)
            if name and position == 0:
                enums.append((name.group('name'), '0'))
            elif name:
                enums.append((name.group('name'),
                             enums[position - 1][0] + ' + 1')) # Prev. var. + 1
            # Offset the increment to position if nothing is being added
            else: position -= 1
        line = next(infile)
        line = remove_comment(line, infile)
        position += 1
    return (enum_group, enums)


def struct_writer(infile):
    """Builds a tuple containing the data to represent structs in python.

    Parameters:
        match: The matched pattern from the regular expression containing the
               name of the current enum.
        infile: The file being read from.

    Returns:
        A tuple in the form (<name of struct>, [list of fields])
    """

    # Maps the data types used in in the header file to the ctypes equivalent.
    types = {'uns32': 'ctypes.c_uint32', 'uns16': 'ctypes.c_uint16',
             'uns64': 'ctypes.c_uint64', 'uns8': 'ctypes.c_uint8',
             'int32': 'ctypes.c_int32', 'int16': 'ctypes.c_int16',
             'int64': 'ctypes.c_int64', 'int8': 'ctypes.c_int8',
             'long64': 'ctypes.c_long', 'char*': 'ctypes.c_char_p',
             'void*': 'ctypes.c_void_p', 'flt64': 'ctypes.c_float',
             'rs_bool': 'ctypes.c_short', 'char': 'ctypes.c_char',
             'void': 'ctypes.c_void_p'}

    fields = []
    # Skip two lines ahead to ignore the '{' line
    next(infile)
    line = next(infile)
    line = remove_comment(line, infile)

    struct_pattern = '^\s+(const )?(?P<type>\w+\*?)\s+(?P<var>\w+)\[?(?P<num>\d+)?\]?;$'

    while '}' not in line:
        match = re.search(struct_pattern, line)
        if match:
            # EAFP
            try:
                var_type = types[match.group('type')]
            except KeyError:
                if match.group('type')[-1] == '*' and match.group('type')[:-1] in types:
                    var_type = types[match.group('type')[:-1]]
                else:
                    var_type = types['void'] # Handle bug for md_ext_item_info
            variable = match.group('var')
            if match.group('num'):
                var_type += ' * {}'.format(match.group('num'))
            if var_type[-1] == '*':
                var_type = 'ctypes.POINTER({})'.format(var_type[:-1])
            fields.append((variable, var_type))
        line = next(infile)
        line = remove_comment(line, infile)

    line = next(infile)
    line = remove_comment(line, infile)
    struct_name = '^(?P<name>\w+);$'
    name = re.search(struct_name, line)
    return name.group('name'), fields


def remove_comment(to_remove, infile):
    """Removes trailing block comments from the end of a string.

    Parameters:
        to_remove: The string to remove the comment from.
        infile: The file being read from.

    Returns:
        The paramter string with the block comment removed (if comment was
        present in string).
    """
    start_comment = re.search('\s*(\/\*|//)', to_remove)

    # Remove comments if they are in the matched group.
    if start_comment:
        end_comment = re.search('.*\*\/', to_remove)
        if end_comment or ('//' in to_remove and not '/*' in to_remove) :
            removed = to_remove[:start_comment.start(0)] + '\n'
            return removed
        while not end_comment:
            to_remove = next(infile)
            end_comment = end_comment = re.search('.*\*\/', to_remove)
        return ''
    else:
        removed = to_remove
    return removed


def parse_line(line, infile, collection):
    """Determines if a line is a define or a start of an enum or struct.

    After determining the structure of a current line, parse_line will then
    call the appropriate funcion in order to deal with the current line.

    Parameters:
        line: The current line being read in from the input file.
        infile: The opened input file.
        collection: The dictionary containing all the lists of data.
    """
    # Define, enum, and struct check pattern.
    define = '^#define (?P<var>\w+)\s+(?P<val>.+)$'
    enum = '^typedef enum (?P<name>\w+)$'
    struct = '^typedef struct .+$'

    expressions = {'defines': define, 'enums': enum, 'structs': struct}

    for key in expressions:
        match = re.search(expressions[key], line)
        if match:
            if key == 'defines':
                collection[key].append(define_writer(match))
            if key == 'enums':
                collection[key].append(enum_writer(match, infile))
            if key == 'structs':
                collection[key].append(struct_writer(infile))




if __name__ == '__main__':
    header_comment = \
"""###############################################################################
# File: constants.py
# Author: Cameron Smith
# Date of Last Edit: {}
#
# Purpose: To maintain the naming conventions used with PVCAM.h for Python
#          scripts.
#
# Notes: This file is generated by constants_generator.py. See that script for
#        details about implementation. Please do not alter this file. Instead,
#        make any additional changes to the constants_generator.py if
#        additional data is needed.
#
# Bugs: [See constants_generator.py]
###############################################################################\n"""\
    .format(date.today())

    defines = []
    macros = []
    enums = []
    structs = []
    collection = {'defines': defines, 'macros': macros, 'enums': enums,
                  'structs': structs}
    # Open file and read parse each line before writing
    try:
        with open(pvcam, 'r') as infile:
            for line in infile:
                line = remove_comment(line, infile)
                parse_line(line, infile, collection)

    except FileNotFoundError:
        print(pvcam + " not found. Make sure that the PVCAM SDK is installed "
              "on this machine.")
        exit(-1)

    with open(constants, 'w') as outfile:
        outfile.write(header_comment)
        outfile.write('import ctypes,os\n')
        outfile.write('### DEFINES ###\n')
        for tup in collection['defines']:
            outfile.write('{} = {}\n'.format(tup[0], tup[1]))
        outfile.write('\n### ENUMS ###\n')
        for tup in collection['enums']:
            outfile.write('\n# {}\n'.format(tup[0]))
            for item in tup[1]:
                outfile.write('{} = {}\n'.format(item[0], item[1]))
        outfile.write('\n### STRUCTS ###\n')
        for tup in collection['structs']:
            outfile.write('class {}(ctypes.Structure):\n'.format(tup[0]))
            outfile.write('    _fields_ = [\n')
            for item in tup[1]:
                outfile.write('                (\'{}\', {}),\n'.format(item[0],
                                                                       item[1]))
            outfile.write('               ]\n\n')

    # Diagnostic
    if False:
        print('Defines: {}'.format(len(collection['defines'])))
        print('Enums: {}'.format(len(collection['enums'])))
        for tup in collection['enums']:
            print('---{}: {}'.format(tup[0].ljust(20), len(tup[1])))
        print('Structs: {}'.format(len(collection['structs'])))
        for item in collection['structs']:
            print('---{}: {}'.format(item[0], item[1]))



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
from queue import Queue
import threading
from copy import deepcopy
import time

# Third Party Imports
import numpy as np
from scipy.optimize import curve_fit

# Local imports
from navigate.model.features.feature_container import load_features
import navigate.model.analysis.image_contrast as img_contrast
from navigate.model.features.image_writer import ImageWriter


def poly2(x, a, b, c):
    """Second order polynomial function

    Parameters
    ----------
    x : float
        x value
    a : float
        a value
    b : float
        b value
    c : float
        c value

    Returns
    -------
    float
        y value
    """
    return a * x**2 + b * x + c


def gauss(x, a, b, c, d):
    """Gaussian function

    Parameters
    ----------
    x : float
        x value
    a : float
        a value
    b : float
        b value
    c : float
        c value
    d : float
        d value

    Returns
    -------
    float
        y value
    """
    y = (x - b) / c
    y = np.exp(-(y**2))
    return a * y + d


def r_squared(y, y_fit):
    """Calculate the R^2 value

    Parameters
    ----------
    y : list
        y values
    y_fit : list
        y_fit values

    Returns
    -------
    float
        R^2 value
    """
    y_bar = np.mean(y)
    SS_res = np.sum((y - y_fit) ** 2)
    SS_tot = np.sum((y - y_bar) ** 2)

    return 1 - SS_res / SS_tot


def fourier_annulus(im, radius_1=0, radius_2=64):
    """Calculate the mean of the fourier transform of an annulus

    Parameters
    ----------
    im : array
        Image array
    radius_1 : int, optional
        Inner radius of the annulus, by default 0
    radius_2 : int, optional
        Outer radius of the annulus, by default 64

    Returns
    -------
    float
        Mean of the fourier transform of the annulus
    array
        Fourier transform of the annulus
    """

    x_, y_ = np.meshgrid(range(im.shape[1]), range(im.shape[0]))

    x_ = x_.astype(float) - x_.mean()
    y_ = y_.astype(float) - y_.mean()

    mask = x_**2 + y_**2 > radius_1**2
    mask *= x_**2 + y_**2 <= radius_2**2

    IM = np.fft.fftshift(np.fft.fft2(im))
    IM_abs = np.abs(IM)

    IM_mask = IM_abs * mask

    return np.mean(IM_mask), IM_mask


class TonyWilson:
    """Tony Wilson iterative AO routine"""

    def __init__(self, model):
        """Initialize the Tony Wilson iterative AO routine

        Parameters
        ----------
        model : Model
            Model object
        """
        #: int: Number of modes
        self.n_modes = None

        #: int: Number of iterations
        self.n_iter = None

        #: int: Number of steps
        self.n_steps = None

        #: float: Coefficient amplitude
        self.coef_amp = None

        #: bool: True if all iterations are done, False otherwise
        self.done_all = False

        #: bool: True if the current iteration is done, False otherwise
        self.done_itr = False

        #: list: detailed report to save as JSON after
        self.report = []

        # TODO: I don't think these are used...
        self.laser = None
        self.laser_power = 0
        self.start_time = 0

        #: navigate.model.Model: Model object
        self.model = model

        #: navigate.model.devices.mirrors.mirror_imop.ImagineOpticsMirror: Mirror object
        self.mirror_controller = self.model.active_microscope.mirror.mirror_controller

        #: int: Number of modes
        self.n_modes = self.mirror_controller.n_modes

        #: bool: whether to save report at the end of run
        self.save_report = self.model.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ].get("save_report", False)

        #: list: List of coefficients to change
        self.change_coef = []
        modes_armed_dict = self.model.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ]["TonyWilson"]["modes_armed"]

        #: list: List of mode names
        self.mode_names = modes_armed_dict.keys()
        for i, k in enumerate(self.mode_names):
            if modes_armed_dict[k]:
                self.change_coef += [i]
        self.n_coefs = len(self.change_coef)

        start_from = self.model.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]["from"]
        if start_from == "flat":
            self.best_coefs = np.zeros(self.n_modes, dtype=np.float32)
        elif start_from == "current":
            curr_expt_coefs = list(
                self.model.configuration["experiment"]["MirrorParameters"][
                    "modes"
                ].values()
            )
            self.best_coefs = np.asarray(curr_expt_coefs, dtype=np.float32)

        self.metric = self.model.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ]["TonyWilson"]["metric"]

        self.fit_func = self.model.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ]["TonyWilson"]["fitfunc"]

        self.best_coefs_overall = deepcopy(self.best_coefs)
        self.best_metric = 0.0
        self.coef_sweep = None
        self.best_peaks = []

        # Queue
        self.tw_frame_queue = Queue()
        self.tw_data_queue = Queue()

        # Image Writer
        self.image_writer = ImageWriter(model, sub_dir="AO_Frames")

        # target channel
        self.target_channel = 1

        self.config_table = {
            "signal": {
                "init": self.pre_func_signal,
                "main": self.in_func_signal,
                "end": self.end_func_signal,
            },
            "data": {
                "init": self.pre_func_data,
                "main": self.in_func_data,
                "end": self.end_func_data,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }

    def run(self, *args):
        """Run the Tony Wilson iterative AO routine

        Parameters
        ----------
        args[0] : dict
            Current microscope state.
        args[1] : dict
            Autofocus parameters

        """
        frame_num = self.get_tw_frame_num()
        if frame_num < 1:
            return

        # Opens correct shutter and puts all signals to false
        self.model.prepare_acquisition()
        self.model.active_microscope.prepare_next_channel()

        # load signal and data containers
        self.model.signal_container, self.model.data_container = load_features(
            self.model, [[{"name": TonyWilson}]]
        )

        self.model.signal_thread = threading.Thread(
            target=self.model.run_acquisition, name="TonyWilson Signal"
        )

        self.model.data_thread = threading.Thread(
            target=self.model.run_data_process,
            # args=(frame_num,),
            kwargs={"data_func": self.image_writer.save_image},
            name="TonyWilson Data",
        )

        print("\n**** STARTING TONY WILSON ****\n")

        # Start Threads
        self.model.signal_thread.start()
        self.model.data_thread.start()

    def get_tw_frame_num(self):
        """Calculate how many frames are needed: iterations x steps x num_coefs"""
        settings = self.model.configuration["experiment"]["AdaptiveOpticsParameters"][
            "TonyWilson"
        ]
        frames = settings["iterations"] * settings["steps"] * self.n_coefs
        return frames

    # don't need this?
    def get_steps(self, ranges, step_size):
        """Calculate the number of steps and the position offset

        Parameters
        ----------
        ranges : int
            Range of the scan
        step_size : int
            Step size

        Returns
        -------
        int
            Number of steps
        int
            Position offset
        """
        steps = ranges // step_size + 1
        pos_offset = (steps // 2) * step_size + step_size
        return steps, pos_offset

    def pre_func_signal(self):
        """Prepare the signal"""
        # initialize the mirror and coef lists, etc

        # Timing
        self.start_time = time.time()

        tw_settings = self.model.configuration["experiment"][
            "AdaptiveOpticsParameters"
        ]["TonyWilson"]

        self.done_all = False

        self.n_iter = tw_settings["iterations"]
        self.n_steps = tw_settings["steps"]
        self.coef_amp = tw_settings["amplitude"]

        self.coef_sweep = np.linspace(
            -self.coef_amp, self.coef_amp, self.n_steps
        ).astype(np.float32)

        self.signal_id = 0
        self.target_signal_id = 0
        self.total_frame_num = self.get_tw_frame_num()

        print(f"Total frame num: {self.total_frame_num}")

    def in_func_signal(self):
        """Run the signal.

        Returns
        -------
        bool
            True if the signal is done, False otherwise
        """
        out_str = "in_func_signal\n"

        out_str += f"\tSignal:\t{self.signal_id}\n"

        step = self.signal_id % self.n_steps
        coef = int(self.signal_id / self.n_steps) % self.n_coefs
        itr = int(self.signal_id / self.n_steps / self.n_coefs) % self.n_iter

        out_str += f"\tStep:\t{step}\n"
        out_str += f"\tC_n:\t{coef}\n"
        out_str += f"\tItr:\t{itr}\n"

        coef_arr = np.zeros(self.n_modes, dtype=np.float32)
        c = self.change_coef[coef]
        coef_arr[c] = self.coef_sweep[step]

        applied_coefs = coef_arr + self.best_coefs

        out_str += f"\tApply:\t[{' '.join([f'{c:.2f}' for c in (applied_coefs)])}]\n"

        # Update the mirror...
        self.mirror_controller.flat()
        try:
            self.mirror_controller.display_modes(applied_coefs)
        except Exception as e:
            print(e)
            return

        time.sleep(self.model.active_microscope.current_exposure_time / 1000)

        try:
            curr_mirror_coefs = self.mirror_controller.get_modal_coefs()[0]
            out_str += (
                f"\tCoefs:\t[{' '.join([f'{c:.2f}' for c in (curr_mirror_coefs)])}]\n"
            )
        except Exception as e:
            print(e)
            return

        self.model.logger.debug(
            f"*** TonyWilson > in_func_signal :: iter:{itr}\tcoef:{coef}\tstep:{step}"
        )
        coef_str = " ".join([f"{c:.2f}" for c in (coef_arr + self.best_coefs)])
        self.model.logger.debug(
            f"*** TonyWilson > in_func_signal :: display_modes:[{coef_str}]"
        )

        if (applied_coefs == curr_mirror_coefs).all() or (applied_coefs == 0).all():
            self.signal_id += 1

            out_str += "\tSending tw_frame_queue...\n"
            self.tw_frame_queue.put(
                (
                    self.model.frame_id,
                    self.total_frame_num - self.signal_id,
                    itr,
                    coef,
                    step,
                )
            )
        else:
            out_str += "\tMirror update failed...\n"

        print(out_str)

        return self.signal_id >= self.total_frame_num

    def end_func_signal(self):
        """End the signal

        Returns
        -------
        bool
            True if the signal is done, False otherwise
        """
        print("end_func_signal() called!!!")

        if self.model.stop_acquisition:
            return True

        return self.signal_id >= self.total_frame_num

    def pre_func_data(self):
        """Prepare the data"""
        self.f_frame_id = (
            -1
        )  # to indicate if there is one frame need to calculate shannon value,
        # but the image frame isn't ready
        self.frame_num = 10  # any value but not 1
        self.target_frame_id = 0  # frame id in the buffer with best focus
        self.get_frames_num = 0
        self.plot_data = []
        self.trace_list = {}
        self.total_frame_num = self.get_tw_frame_num()
        self.x = self.coef_sweep
        self.y = []
        self.x_fit = np.linspace(-self.coef_amp, self.coef_amp, 1024)
        self.y_fit = []
        self.mirror_img = None

        self.frames_done = 0

    def process_data(self, coef, mode="poly"):
        """Process the data

        Parameters
        ----------
        coef : int
            Coefficient index
        mode : str, optional
            Fitting mode, by default "poly"
        """
        self.y = self.plot_data

        if mode == "poly":
            c = np.min(self.y)  # offset guess
            b = (np.max(self.y) - c) / self.coef_amp  # slope guess
            a = -b / 2

            p, _ = curve_fit(
                poly2,
                self.x,
                self.y,
                p0=[a, b, c],
                bounds=([-np.inf, -np.inf, -np.inf], [0.0, np.inf, np.inf]),
            )
            self.y_fit = poly2(self.x_fit, p[0], p[1], p[2])
            r_2 = r_squared(self.y, poly2(self.x, p[0], p[1], p[2]))

        elif mode == "gauss":
            d = np.min(self.y)
            a = np.max(self.y) - d
            b = self.x[np.argmax(self.y)]
            c = (self.x[-1] - self.x[0]) / 2

            p, _ = curve_fit(
                gauss,
                self.x,
                self.y,
                p0=[a, b, c, d],
                bounds=([0, -np.inf, 0, 0], [np.inf, np.inf, np.inf, np.inf]),
            )
            self.y_fit = gauss(self.x_fit, p[0], p[1], p[2], p[3])
            r_2 = r_squared(self.y, gauss(self.x, p[0], p[1], p[2], p[3]))

        self.best_coefs[self.change_coef[coef - 1]] += (
            self.x_fit[self.y_fit.argmax()] * r_2
        )  # weight by R^2 goodness of fit
        self.mirror_img = self.mirror_controller.get_wavefront_pix()

        new_metric = self.plot_data[int(self.n_steps / 2)]
        self.best_peaks.append(new_metric)
        if new_metric > self.best_metric:
            self.best_metric = new_metric
            self.best_coefs_overall = deepcopy(self.best_coefs)

        self.plot_data = []

    def in_func_data(self, frame_ids=[]):
        """Run the data

        Parameters
        ----------
        frame_ids : list, optional
            List of frame ids, by default []

        Returns
        -------
        list
            List of frame ids
        """
        self.get_frames_num += len(frame_ids)

        out_str = "in_func_data\n"

        out_str += f"\tFrames:\t{frame_ids}\n"

        out_str += f"\tGet Frames Num:\t{self.get_frames_num}\n"

        while True:
            try:
                if self.f_frame_id < 0:
                    (
                        self.f_frame_id,
                        self.frame_num,
                        itr,
                        coef,
                        step,
                    ) = self.tw_frame_queue.get_nowait()
                if self.f_frame_id not in frame_ids:
                    out_str += (
                        f"\tFrame ID {self.f_frame_id} was not in the frame queue...\n"
                    )
                    break
            except Exception:
                out_str += "\tStill waiting tw_frame_queue...\n"
                break

            # get the image metric
            img = self.model.data_buffer[self.f_frame_id]

            """ IMAGE METRICS """
            if self.metric == "Pixel Max":
                new_data = img.max()
            elif self.metric == "Pixel Average":
                new_data = img.mean()
            elif self.metric == "DCT Shannon Entropy":
                new_data = img_contrast.fast_normalized_dct_shannon_entropy(img, 3)[0]

            if len(self.plot_data) == self.n_steps:
                self.process_data(coef, mode=self.fit_func)
                self.trace_list[self.mode_names[self.change_coef[coef]]] = {
                    "x": self.x,
                    "y": self.y,
                    "x_fit": self.x_fit[::32],
                    "y_fit": self.y_fit[::32],
                }
                out_str += "\tFITTING DATA...\n"

            self.plot_data.append(new_data)
            out_str += f"\tTrace:\t{np.flip(self.plot_data)}\n"

            self.frames_done += 1
            out_str += f"\tDone:\t{self.frames_done}\n"

            self.f_frame_id = -1

            self.model.logger.debug(
                f"*** TonyWilson > in_func_data :: plot_data: {np.flip(self.plot_data)}"
            )

            out_str += f"\tFrame Num:\t{self.frame_num}\n"
            if self.frame_num == 1:
                self.frame_num = 10  # any value but not 1
                return [self.target_frame_id]

            if coef == self.n_coefs - 1:
                if step == self.n_steps - 1:

                    self.coef_sweep *= 0.95
                    self.done_itr = True
                    out_str += f"\tDone iteration {itr}!\n"

                    if itr == self.n_iter - 1:
                        self.done_all = True
                        out_str += "\tDone all!!!\n"

        out_str += "\tSending tw_data_queue...\n"
        self.tw_data_queue.put((self.frames_done,))

        print(out_str)

        if self.frames_done >= self.total_frame_num:
            print(">>> in_func_data ended!!!")
            return frame_ids

    def build_report(self):

        return deepcopy(
            {
                "mirror_update": {
                    "mirror_img": self.mirror_img,
                    "coefs": self.best_coefs,
                },
                "tonywilson": {
                    "peaks": self.best_peaks,
                    "trace": self.trace_list,
                    "done": self.done_all,
                    "metric": self.metric,
                    "iter": self.n_iter,
                    "steps": self.n_steps,
                    "amp": self.coef_amp,
                    "modes": self.change_coef,
                },
            }
        )

    def end_func_data(self):
        """End the data

        Returns
        -------
        bool
            True if the data is done, False otherwise
        """
        print("end_func_data() called!!!")

        if self.done_all:
            self.best_coefs = self.best_coefs_overall
            self.model.stop_acquisition = True
            self.model.end_acquisition()
            print("Ending acquisition...")
            try:
                stop_time = time.time()
                print(f"Total runtime:\t{(stop_time - self.start_time):.3f} sec")
            except Exception as e:
                print(e)

        try:
            if self.done_itr:
                current_report = self.build_report()
                self.report.append(current_report)

                self.model.event_queue.put(
                    ("mirror_update", current_report["mirror_update"])
                )
                self.model.event_queue.put(("tonywilson", current_report["tonywilson"]))
        except Exception as e:
            print(e)

        self.done_itr = False

        self.mirror_controller.display_modes(self.best_coefs_overall)

        if self.done_all and self.save_report:
            self.model.event_queue.put(("ao_save_report", self.report))

        return self.frames_done >= self.total_frame_num



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time

# Third Party Imports

# Local Imports
from navigate.model.devices.filter_wheel.filter_wheel_base import FilterWheelBase
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_filter_wheel_connection(comport, baudrate=115200, timeout=0.25):
    """Build ASIFilterWheel Serial Port connection

    Parameters
    ----------
    comport : str
        Comport for communicating with the filter wheel, e.g., COM1.
    baudrate : int
        Baud rate for communicating with the filter wheel, default is 115200.
    timeout : float
        Timeout for communicating with the filter wheel, default is 0.25.

    Returns
    -------
    tiger_controller : TigerController
        ASI Tiger Controller object.
    """
    # wait until ASI device is ready
    tiger_controller = TigerController(comport, baudrate)
    tiger_controller.connect_to_serial()
    if not tiger_controller.is_open():
        raise Exception("ASI stage connection failed.")
    return tiger_controller


class ASIFilterWheel(FilterWheelBase):
    """ASIFilterWheel - Class for controlling ASI Filter Wheels

    Note
    ----
        Additional information on the ASI Filter Wheel can be found at:
        https://asiimaging.com/docs/fw_1000#fw-1000_ascii_command_set
    """

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize the ASIFilterWheel class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : dict
            Dictionary of device connections.
        configuration : dict
            Dictionary of configuration parameters.
        """

        super().__init__(microscope_name, device_connection, configuration)

        #: obj: ASI Tiger Controller object.
        self.filter_wheel = device_connection

        #: str: Name of the ASI Filter Wheel.
        self.microscope_name = microscope_name

        #: int: Number of filter wheels.
        self.number_of_filter_wheels = configuration["configuration"]["microscopes"][
            microscope_name
        ]["filter_wheel"]["hardware"]["wheel_number"]

        #: float: Delay for filter wheel to change positions.
        self.wait_until_done_delay = configuration["configuration"]["microscopes"][
            microscope_name
        ]["filter_wheel"]["filter_wheel_delay"]

        # Send Filter Wheel/Wheels to Zeroth Position
        for i in range(self.number_of_filter_wheels):
            self.filter_wheel.select_filter_wheel(filter_wheel_number=i)

            #: int: Active filter wheel.
            self.active_filter_wheel = i
            self.filter_wheel.move_filter_wheel(filter_wheel_position=0)

            #: int: Filter wheel position.
            self.filter_wheel_position = 0

    def __enter__(self):
        """Enter the ASI Filter Wheel context manager."""
        return self

    def __exit__(self):
        """Exit the ASI Filter Wheel context manager."""
        if self.filter_wheel.is_open():
            logger.debug("ASI Filter Wheel - Closing Device.")
            self.filter_wheel.disconnect_from_serial()

    def filter_change_delay(self, filter_name):
        """Estimate duration of time necessary to move the filter wheel

        Assumes that it is ~40ms per adjacent position.
        Depends on filter wheel parameters and load.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.

        """
        old_position = self.filter_wheel_position
        new_position = self.filter_dictionary[filter_name]
        delta_position = int(abs(old_position - new_position))
        self.wait_until_done_delay = delta_position * 0.04

    def set_filter(self, filter_name, wait_until_done=True):
        """Change the filter wheel to the filter designated by the filter
        position argument.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        if self.check_if_filter_in_filter_dictionary(filter_name) is True:

            # Calculate the Delay Needed to Change the Positions
            self.filter_change_delay(filter_name)

            for wheel_idx in range(self.number_of_filter_wheels):
                self.filter_wheel.select_filter_wheel(filter_wheel_number=wheel_idx)
                self.filter_wheel.move_filter_wheel(self.filter_dictionary[filter_name])

            #  Wheel Position Change Delay
            if wait_until_done:
                time.sleep(self.wait_until_done_delay)

    def close(self):
        """Close the ASI Filter Wheel serial port.

        Sets the filter wheel to the home position and then closes the port.
        """
        if self.filter_wheel.is_open():
            self.filter_wheel.move_filter_wheel_to_home()
            logger.debug("ASI Filter Wheel - Closing Device.")
            self.filter_wheel.disconnect_from_serial()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import tkinter

# Third Party Imports

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.view.popups.microscope_setting_popup_window import (
    MicroscopeSettingPopupWindow,
)

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MicroscopePopupController(GUIController):
    """Controller for the microscope setting popup window."""

    def __init__(self, view, parent_controller, microscope_info):
        """Initialize the MicroscopePopupController.

        Parameters
        ----------
        view : object
            GUI element containing widgets and variables to control.
            Likely tk.Toplevel-derived.
        parent_controller : Navigate_controller
            The main controller.
        microscope_info : dict
            A dictionary containing the information of the microscopes.

        """
        super().__init__(view, parent_controller)
        #: tk.Toplevel-derived: The popup window.
        self.view = MicroscopeSettingPopupWindow(view, microscope_info)
        #: dict: A dictionary containing the widgets.
        self.widgets = self.view.get_widgets()
        #: dict: A dictionary containing the variables.
        self.variables = self.view.get_variables()
        #: dict: A dictionary containing the buttons.
        self.buttons = self.view.get_buttons()
        #: dict: A dictionary containing the microscope information.
        self.microscope_info = microscope_info
        #: str: The name of the primary microscope.
        self.primary_microscope = None

        # add saving function to the function closing the window
        self.view.popup.protocol(
            "WM_DELETE_WINDOW",
            self.exit_func,
        )

        for microscope_name in self.microscope_info:
            self.widgets[microscope_name].widget["values"] = (
                "Primary Microscope",
                "Additional Microscope",
                "Not Use",
            )
            self.widgets[microscope_name].widget.bind(
                "<<ComboboxSelected>>", self.update_microscope_mode(microscope_name)
            )

        # button events
        self.buttons["confirm"].configure(command=self.confirm_microscope_setting)
        self.buttons["cancel"].configure(command=self.exit_func)

    def showup(self):
        """This function will let the popup window show in front."""
        self.view.popup.deiconify()
        # self.view.popup.attributes("-topmost", 1)

    def exit_func(self):
        """Close the window and delete this controller."""
        self.view.popup.dismiss()
        delattr(self.parent_controller, "microscope_popup_controller")

    def update_microscope_mode(self, microscope_name):
        """Update the microscope mode.

        Parameters
        ----------
        microscope_name : str
            The name of the microscope.

        Returns
        -------
        func
            A function to update the microscope mode.
        """

        def clear_device_info(microscope_name):
            for microscope in self.microscope_info.keys():
                if microscope != microscope_name:
                    for k in self.microscope_info[microscope_name].keys():
                        if (
                            self.microscope_info[microscope_name][k]
                            == self.microscope_info[microscope][k]
                        ):
                            self.variables[f"{microscope_name} {k}"].set("")

        variable = self.variables[microscope_name]
        microscope_name = microscope_name

        def func(*args):
            """Update the microscope mode.

            Parameters
            ----------
            *args : list
                The arguments.
            """
            setting_value = variable.get()
            if setting_value == "Primary Microscope":
                self.primary_microscope = microscope_name
                # check the status of other microscopes
                for microscope in self.microscope_info.keys():
                    if (
                        microscope != microscope_name
                        and self.variables[microscope].get() == setting_value
                    ):
                        self.variables[microscope].set("Additional Microscope")
                        clear_device_info(microscope)
            elif self.primary_microscope == microscope_name:
                self.primary_microscope = None

            if setting_value == "Additional Microscope":
                clear_device_info(microscope_name)
            else:
                # set all the devices to real device
                for k in self.microscope_info[microscope_name]:
                    self.variables[f"{microscope_name} {k}"].set(
                        self.microscope_info[microscope_name][k]
                    )

        return func

    def confirm_microscope_setting(self):
        """Confirm the microscope settings."""
        # must have one primary microscope
        if not self.primary_microscope:
            tkinter.messagebox.showerror(
                title="Warning",
                message="There is no Primary Microscope! " "Please select one!",
            )
            self.showup()
            return
        # tell the controller to start a new Camera View Window and prepare a virtual
        # microscope
        if (
            self.primary_microscope
            != self.parent_controller.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ]
        ):
            self.parent_controller.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ] = self.primary_microscope
            zoom = self.parent_controller.configuration["waveform_constants"][
                "remote_focus_constants"
            ][self.primary_microscope].keys()[0]
            self.parent_controller.configuration["experiment"]["MicroscopeState"][
                "zoom"
            ] = zoom
            self.parent_controller.resolution_value.set(
                f"{self.primary_microscope} {zoom}"
            )

        has_multi_cameras = False
        for microscope_name in self.microscope_info.keys():
            if self.variables[microscope_name].get() == "Additional Microscope":
                config = {}
                for k in self.microscope_info[microscope_name].keys():
                    config[k] = self.variables[f"{microscope_name} {k}"].get()
                self.parent_controller.additional_microscopes_configs[
                    microscope_name
                ] = config
                has_multi_cameras = True
            elif (
                microscope_name in self.parent_controller.additional_microscopes_configs
            ):
                del self.parent_controller.additional_microscopes_configs[
                    microscope_name
                ]

        # enable/disable change 'resolution'(microscope) menu if there are multiple
        # cameras running
        # enable primary camera
        for microscope_name in self.microscope_info.keys():
            if microscope_name == self.primary_microscope or not has_multi_cameras:
                # enable
                self.parent_controller.view.menubar.menu_resolution.entryconfig(
                    microscope_name, state="normal"
                )
            else:
                # disable
                self.parent_controller.view.menubar.menu_resolution.entryconfig(
                    microscope_name, state="disabled"
                )

        self.exit_func()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library imports
import logging
import importlib  # noqa: F401
from multiprocessing.managers import ListProxy

from navigate.model.device_startup_functions import (
    start_stage,
)
from navigate.tools.common_functions import build_ref_name
from navigate.model.devices.stages.stage_galvo import GalvoNIStage

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Microscope:
    """Microscope Class - Used to control the microscope."""

    def __init__(
        self, name, configuration, devices_dict, is_synthetic=False, is_virtual=False
    ):
        """Initialize the microscope.

        Parameters
        ----------
        name : str
            Name of the microscope.
        configuration : dict
            Configuration dictionary.
        devices_dict : dict
            Dictionary of devices.
        is_synthetic : bool, optional
            Is synthetic, by default False
        is_virtual : bool, optional
            Is virtual, by default False
        """

        # Initialize microscope object
        #: str: Name of the microscope.
        self.microscope_name = name
        #: dict: Configuration dictionary.
        self.configuration = configuration
        #: SharedNDArray: Buffer for image data.
        self.data_buffer = None
        #: dict: Dictionary of stages.
        self.stages = {}
        #: list: List of stages.
        self.stages_list = []
        #: bool: Ask stage for position.
        self.ask_stage_for_position = True
        #: dict: Dictionary of lasers.
        self.lasers = {}
        #: dict: Dictionary of galvanometers.
        self.galvo = {}
        #: dict: Dictionary of data acquisition devices.
        self.daq = devices_dict.get("daq", None)
        #: object; Tiger Controller object.
        self.tiger_controller = None
        #: dict: Dictionary of microscope info.
        self.info = {}
        #: int: Current channel.
        self.current_channel = None
        #: int: Current laser index.
        self.current_laser_index = 0
        #: list: List of all channels.
        self.channels = None
        #: list: List of available channels.
        self.available_channels = None
        #: int: Number of images.
        self.number_of_frames = None
        #: float: Central focus position.
        self.central_focus = None
        #: Bool: Is a synthetic microscope.
        self.is_synthetic = is_synthetic
        #: list: List of laser wavelengths.
        self.laser_wavelength = []
        #: dict: Dictionary of returned stage positions.
        self.ret_pos_dict = {}
        #: dict: Dictionary of commands
        self.commands = {}
        #: dict: Dictionary of plugin devices
        self.plugin_devices = {}

        if is_virtual:
            return

        device_ref_dict = {
            "camera": ["type", "serial_number"],
            "filter_wheel": ["type"],
            "zoom": ["type", "servo_id"],
            "shutter": ["type", "channel"],
            "remote_focus_device": ["type", "channel"],
            "galvo": ["type", "channel"],
            "lasers": ["wavelength"],
            "mirror": ["type"],
        }

        # TODO: This cannot be pulled into the repo.
        #  I need help comping up with a more general way to have
        # shared devices.
        # devices_dict['filter_wheel']['ASI'] = devices_dict['stages']['ASI_119060508']

        device_name_dict = {"lasers": "wavelength"}

        laser_list = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["lasers"]
        self.laser_wavelength = [laser["wavelength"] for laser in laser_list]

        if "__plugins__" not in devices_dict:
            devices_dict["__plugins__"] = {}

        # LOAD/START CAMERAS, FILTER_WHEELS, ZOOM, SHUTTERS, REMOTE_FOCUS_DEVICES,
        # GALVOS, AND LASERS
        for device_name in self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ].keys():
            if device_name in ["daq", "stage"]:
                continue
            is_plugin = False
            device_connection = None
            (
                device_config_list,
                device_name_list,
                is_list,
            ) = self.assemble_device_config_lists(
                device_name=device_name, device_name_dict=device_name_dict
            )
            if device_name not in device_ref_dict:
                if device_name in devices_dict["__plugins__"]:
                    device_ref_dict[device_name] = devices_dict["__plugins__"][
                        device_name
                    ]["ref_list"]
                    is_plugin = True
                else:
                    print(
                        f"Device {device_name} could not be loaded! Please make sure there is no spelling error!"
                    )
                    logger.debug(
                        f"Device {device_name} could not be loaded! Please make sure there is no spelling error!"
                    )
                    continue
            for i, device in enumerate(device_config_list):
                device_ref_name = None
                if "hardware" in device.keys():
                    ref_list = [
                        device["hardware"][k] for k in device_ref_dict[device_name]
                    ]

                else:
                    try:
                        ref_list = [device[k] for k in device_ref_dict[device_name]]
                    except KeyError:
                        ref_list = []

                device_ref_name = build_ref_name("_", *ref_list)
                if (
                    device_name in devices_dict
                    and device_ref_name in devices_dict[device_name]
                ):
                    device_connection = devices_dict[device_name][device_ref_name]

                elif is_plugin:
                    device_plugin_dict = devices_dict.get(device_name, {})
                    try:
                        exec(
                            f"device_plugin_dict['{device_ref_name}'] = devices_dict['__plugins__']['{device_name}']['load_device'](configuration, is_synthetic)"
                        )
                        devices_dict[device_name] = device_plugin_dict
                        device_connection = device_plugin_dict[device_ref_name]
                        exec(
                            f"self.plugin_devices['{device_name}'] = devices_dict['__plugins__']['{device_name}']['start_device'](self.microscope_name, device_connection, configuration, is_synthetic)"
                        )
                    except RuntimeError:
                        print(
                            f"Device {device_name} isn't loaded correctly! Please check the spelling and the plugin!"
                        )
                        continue

                    self.info[device_name] = device_ref_name
                    commands_dict = self.plugin_devices[device_name].commands
                    for command in commands_dict:
                        self.commands[command] = (device_name, commands_dict[command])
                    continue

                # SHARED DEVICES
                elif device_ref_name.startswith("NI") and (
                    device_name == "galvo" or device_name == "remote_focus_device"
                ):
                    device_connection = self.daq

                if device_ref_name.startswith("EquipmentSolutions"):
                    device_connection = self.daq

                if device_ref_name.startswith("ASI") and self.tiger_controller is None:
                    # The first ASI instance of a device connection will be passed to
                    # all other ASI devices as self.tiger_controller
                    device_connection = devices_dict[device_name][device_ref_name]
                    self.tiger_controller = device_connection
                elif (
                    device_ref_name.startswith("ASI")
                    and self.tiger_controller is not None
                ):
                    # If subsequent ASI-based tiger controller devices are included.
                    device_connection = self.tiger_controller

                # LOAD AND START DEVICES
                self.load_and_start_devices(
                    device_name=device_name,
                    is_list=is_list,
                    device_name_list=device_name_list,
                    device_ref_name=device_ref_name,
                    device_connection=device_connection,
                    name=name,
                    i=i,
                    plugin_devices=devices_dict["__plugins__"],
                )

                if device_connection is None and device_ref_name is not None:
                    if device_name not in devices_dict:
                        devices_dict[device_name] = {}

                    devices_dict[device_name][device_ref_name] = (
                        getattr(self, device_name)[device_name_list[i]]
                        if is_list
                        else getattr(self, device_name)
                    )

        # stages
        stage_devices = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["stage"]["hardware"]
        # set the NI Galvo stage flag
        self.configuration["configuration"]["microscopes"][self.microscope_name][
            "stage"
        ]["has_ni_galvo_stage"] = False
        if type(stage_devices) != ListProxy:
            stage_devices = [stage_devices]

        for i, device_config in enumerate(stage_devices):
            device_ref_name = build_ref_name(
                "_", device_config["type"], device_config["serial_number"]
            )

            if device_ref_name not in devices_dict["stages"]:
                logger.debug(f"{device_ref_name} stage not found in the devices_dict!")
                raise Exception(
                    "Stage not found. "
                    "This often arises when the configuration.yaml file is "
                    "incorrectly specified. Please check the "
                    "configuration.yaml file and try again. Things to "
                    "check include: ",
                    "1. If no stage is physically connected to the "
                    "microscope, one must still be listed in the "
                    "configuration.yaml file. We recommend that you use a"
                    "`SyntheticStage`. "
                    "2. If a stage is physically connected to the "
                    "microscope, make sure that it is corrected defined in "
                    "both the `hardware` and `microscopes` sections of the "
                    "configuration.yaml file. Importantly, name, "
                    "type, and serial numbers must match. "
                    "3. You may be using a stage that is not supported by "
                    "Navigate. Please check the list of supported stages in "
                    "the documentation."
                    f"The stage that failed to load is: {device_ref_name}"
                )

            # SHARED DEVICES
            if device_ref_name.startswith("GalvoNIStage"):
                devices_dict["stages"][device_ref_name] = self.daq
                # set the NI Galvo stage flag
                self.configuration["configuration"]["microscopes"][
                    self.microscope_name
                ]["stage"]["has_ni_galvo_stage"] = True

            if device_ref_name.startswith("ASI") and self.tiger_controller is not None:
                # If the self.tiger_controller is already set, then we can pass it to
                # other devices that are connected to the same controller.
                devices_dict["stages"][device_ref_name] = self.tiger_controller

            stage = start_stage(
                microscope_name=self.microscope_name,
                device_connection=devices_dict["stages"][device_ref_name],
                configuration=self.configuration,
                id=i,
                is_synthetic=is_synthetic,
                plugin_devices=devices_dict["__plugins__"],
            )
            for axis in device_config["axes"]:
                self.stages[axis] = stage
                self.info[f"stage_{axis}"] = device_ref_name

            self.stages_list.append((stage, list(device_config["axes"])))

        # connect daq and camera in synthetic mode
        if is_synthetic:
            self.daq.add_camera(self.microscope_name, self.camera)

    def update_data_buffer(self, img_width, img_height, data_buffer, number_of_frames):
        """Update the data buffer for the camera.

        Parameters
        ----------
        img_width : int
            Width of the image.
        img_height : int
            Height of the image.
        data_buffer : numpy.ndarray
            Data buffer for the camera.
        number_of_frames : int
            Number of frames to be acquired.
        """

        if self.camera.is_acquiring:
            self.camera.close_image_series()
        self.camera.set_ROI(img_height, img_width)
        self.data_buffer = data_buffer
        self.number_of_frames = number_of_frames

    def move_stage_offset(self, former_microscope=None):
        """Move the stage to the offset position.

        Parameters
        ----------
        former_microscope : str
            Name of the former microscope.
        """

        if former_microscope:
            former_offset_dict = self.configuration["configuration"]["microscopes"][
                former_microscope
            ]["stage"]
        else:
            former_offset_dict = dict((f"{a}_offset", 0) for a in self.stages)
        self_offset_dict = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["stage"]
        self.ask_stage_for_position = True
        # print(self.stages)
        pos_dict = self.get_stage_position()
        for stage, axes in self.stages_list:
            pos = {
                axis
                + "_abs": (
                    pos_dict[axis + "_pos"]
                    + self_offset_dict[axis + "_offset"]
                    - former_offset_dict[axis + "_offset"]
                )
                for axis in axes
            }
            stage.move_absolute(pos, wait_until_done=True)
        self.ask_stage_for_position = True

    def prepare_acquisition(self):
        """Prepare the acquisition.

        Returns
        -------
        waveform : dict
            Dictionary of all the waveforms.
        """
        self.current_channel = 0
        self.central_focus = None
        self.channels = self.configuration["experiment"]["MicroscopeState"]["channels"]
        self.available_channels = list(
            map(
                lambda c: int(c[len("channel_") :]),
                filter(lambda k: self.channels[k]["is_selected"], self.channels.keys()),
            )
        )
        if self.camera.is_acquiring:
            self.camera.close_image_series()

        # Set Camera Sensor Mode - Must be done before camera is initialized.
        sensor_mode = self.configuration["experiment"]["CameraParameters"][
            "sensor_mode"
        ]
        self.camera.set_sensor_mode(sensor_mode)
        if sensor_mode == "Light-Sheet":
            self.camera.set_readout_direction(
                self.configuration["experiment"]["CameraParameters"][
                    "readout_direction"
                ]
            )
        # set binning
        self.camera.set_binning(
            self.configuration["experiment"]["CameraParameters"]["binning"]
        )
        # Initialize Image Series - Attaches camera buffer and start imaging
        self.camera.initialize_image_series(self.data_buffer, self.number_of_frames)

        # calculate all the waveform
        self.shutter.open_shutter()
        return self.calculate_all_waveform()

    def end_acquisition(self):
        """End the acquisition."""
        self.daq.stop_acquisition()
        # set NI Galvo stage to normal stage mode
        if self.configuration["experiment"]["MicroscopeState"]["image_mode"] == "confocal-projection":
            for stage, _ in self.stages_list:
                if type(stage) == GalvoNIStage:
                    stage.switch_mode("normal")
        self.stop_stage()
        if self.central_focus is not None:
            self.move_stage({"f_abs": self.central_focus})
        if self.camera.is_acquiring:
            self.camera.close_image_series()
        self.shutter.close_shutter()
        for k in self.lasers:
            self.lasers[k].turn_off()
        self.current_channel = 0
        self.central_focus = None

    def turn_on_laser(self):
        """Turn on the current laser."""
        self.lasers[str(self.laser_wavelength[self.current_laser_index])].turn_on()

    def turn_off_lasers(self):
        """Turn off current laser."""
        self.lasers[str(self.laser_wavelength[self.current_laser_index])].turn_off()

    def calculate_all_waveform(self):
        """Calculate all the waveforms.

        Returns
        -------
        waveform : dict
            Dictionary of all the waveforms.
        """
        readout_time = self.get_readout_time()
        exposure_times, sweep_times = self.calculate_exposure_sweep_times(readout_time)
        camera_waveform = self.daq.calculate_all_waveforms(
            self.microscope_name, exposure_times, sweep_times
        )
        remote_focus_waveform = self.remote_focus_device.adjust(
            exposure_times, sweep_times
        )
        galvo_waveform = [
            self.galvo[k].adjust(exposure_times, sweep_times) for k in self.galvo
        ]

        # calculate waveform for galvo stage
        for stage, _ in self.stages_list:
            if type(stage) == GalvoNIStage:
                stage.switch_mode("normal", exposure_times, sweep_times)
        waveform_dict = {
            "camera_waveform": camera_waveform,
            "remote_focus_waveform": remote_focus_waveform,
            "galvo_waveform": galvo_waveform,
        }
        return waveform_dict

    def calculate_exposure_sweep_times(self, readout_time):
        """Calculate the exposure and sweep times for all channels.

        The `calculate_exposure_sweep_times` function calculates and returns exposure
        times and sweep times for all channels in a microscope configuration. It takes
        the camera's readout time as an input parameter and considers various parameters
        such as camera exposure time, delay percentages, and smoothing to compute these
        times. The function iterates through the channels, performs calculations, and
        returns the results as dictionaries containing exposure times and sweep times
        for each channel.

        Parameters
        ----------
        readout_time : float
            Readout time of the camera (seconds) if we are operating the camera in
            Normal mode, otherwise -1.

        Returns
        -------
        exposure_times : dict
            Dictionary of exposure times.
        sweep_times : dict
            Dictionary of sweep times.

        """
        exposure_times = {}
        sweep_times = {}
        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        zoom = microscope_state["zoom"]
        waveform_constants = self.configuration["waveform_constants"]
        camera_delay_percent = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["camera"]["delay_percent"]
        remote_focus_ramp_falling = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["remote_focus_device"]["ramp_falling_percent"]

        duty_cycle_wait_duration = (
            float(
                self.configuration["waveform_constants"]
                .get("other_constants", {})
                .get("remote_focus_settle_duration", 0)
            )
            / 1000
        )
        for channel_key in microscope_state["channels"].keys():
            channel = microscope_state["channels"][channel_key]
            if channel["is_selected"] is True:
                exposure_time = float(channel["camera_exposure_time"]) / 1000

                sweep_time = (
                    exposure_time
                    + exposure_time
                    * (camera_delay_percent + remote_focus_ramp_falling)
                    / 100
                )
                if readout_time > 0:
                    # This addresses the dovetail nature of the camera readout in normal
                    # mode. The camera reads middle out, and the delay in start of the
                    # last lines compared to the first lines causes the exposure to be
                    # net longer than exposure_time. This helps the galvo keep sweeping
                    # for the full camera exposure time.
                    sweep_time += readout_time  # we could set it to 0.14 instead of
                    # 0.16384 according to the test

                ps = float(
                    waveform_constants["remote_focus_constants"][self.microscope_name][
                        zoom
                    ][channel["laser"]].get("percent_smoothing", 0.0)
                )
                if ps > 0:
                    sweep_time = (1 + ps / 100) * sweep_time

                sweep_time += duty_cycle_wait_duration

                exposure_times[channel_key] = exposure_time
                sweep_times[channel_key] = sweep_time

        self.exposure_times = exposure_times
        self.sweep_times = sweep_times

        return exposure_times, sweep_times

    def get_exposure_sweep_times(self):
        """Get the exposure and sweep times for all channels.

        Returns
        -------
        exposure_times : dict
            Dictionary of exposure times.
        sweep_times : dict
            Dictionary of sweep times.

        """
        return self.exposure_times, self.sweep_times

    def prepare_next_channel(self, update_daq_task_flag=True):
        """Prepare the next channel.

        This function, `prepare_next_channel`, is responsible for configuring various
        hardware components for the next imaging channel in an experimental setup.
        It sequentially selects the next available channel, sets the filter wheel,
        camera exposure time, laser power, and other parameters based on the selected
        channel's configuration. Additionally, it stops data acquisition, prepares the
        data acquisition system for the new channel, and adjusts the focus position as
        necessary, ensuring the hardware is ready for imaging the selected channel.

        Parameters
        ----------
        update_daq_task_flag : bool
            whether to override waveforms in the DAQ (create new tasks)
        """
        curr_channel = self.current_channel
        prefix = "channel_"
        if self.current_channel == 0:
            self.current_channel = self.available_channels[0]
        else:
            idx = (self.available_channels.index(self.current_channel) + 1) % len(
                self.available_channels
            )
            self.current_channel = self.available_channels[idx]
        if curr_channel == self.current_channel:
            return

        channel_key = prefix + str(self.current_channel)
        channel = self.configuration["experiment"]["MicroscopeState"]["channels"][
            channel_key
        ]
        # Filter Wheel Settings.
        self.filter_wheel.set_filter(channel["filter"])

        # Camera Settings
        self.current_exposure_time = float(channel["camera_exposure_time"])
        if (
            self.configuration["experiment"]["CameraParameters"]["sensor_mode"]
            == "Light-Sheet"
        ):
            (
                self.current_exposure_time,
                camera_line_interval,
            ) = self.camera.calculate_light_sheet_exposure_time(
                self.current_exposure_time,
                int(
                    self.configuration["experiment"]["CameraParameters"][
                        "number_of_pixels"
                    ]
                ),
            )
            self.camera.set_line_interval(camera_line_interval)
        self.camera.set_exposure_time(self.current_exposure_time)

        # Laser Settings
        self.current_laser_index = channel["laser_index"]
        for k in self.lasers:
            self.lasers[k].turn_off()
        self.lasers[str(self.laser_wavelength[self.current_laser_index])].set_power(
            channel["laser_power"]
        )
        # self.lasers[str(self.laser_wavelength[self.current_laser_index])].turn_on()

        # stop daq before writing new waveform
        # When called the first time, throws an error.
        # choose to not update the waveform is very useful when running ZStack
        # if there is a NI Galvo stage in the system.
        if update_daq_task_flag:
            self.daq.stop_acquisition()
            self.daq.prepare_acquisition(channel_key)

        # Add Defocus term
        # Assume wherever we start is the central focus
        # TODO: is this the correct assumption?
        if self.central_focus is None:
            self.central_focus = self.get_stage_position().get("f_pos")
        if self.central_focus is not None:
            self.move_stage(
                {"f_abs": self.central_focus + float(channel["defocus"])},
                wait_until_done=True,
                update_focus=False,
            )

    def get_readout_time(self):
        """Get readout time from camera.

        Get the camera readout time if we are in normal mode.
        Return a -1 to indicate when we are not in normal mode.
        This is needed in daq.calculate_all_waveforms()

        Returns
        -------
        readout_time : float
            Camera readout time in seconds or -1 if not in Normal mode.
        """
        readout_time = 0
        if (
            self.configuration["experiment"]["CameraParameters"]["sensor_mode"]
            == "Normal"
        ):
            readout_time, _ = self.camera.calculate_readout_time()
        return readout_time

    def move_stage(self, pos_dict, wait_until_done=False, update_focus=True):
        """Move stage to a position.

        Parameters
        ----------
        pos_dict : dict
            Dictionary of stage positions.
        wait_until_done : bool, optional
            Wait until stage is done moving, by default False
        update_focus : bool, optional
            Update the central focus

        Returns
        -------
        success : bool
            True if stage is successfully moved, False otherwise.
        """
        self.ask_stage_for_position = True

        if len(pos_dict.keys()) == 1:
            axis_key = list(pos_dict.keys())[0]
            axis = axis_key[: axis_key.index("_")]
            if update_focus and axis == "f":
                self.central_focus = None
            return self.stages[axis].move_axis_absolute(
                axis, pos_dict[axis_key], wait_until_done
            )

        success = True
        for stage, axes in self.stages_list:
            pos = {
                axis: pos_dict[axis]
                for axis in pos_dict
                if axis[: axis.index("_")] in axes
            }
            if pos:
                success = stage.move_absolute(pos, wait_until_done) and success

        if update_focus and "f_abs" in pos_dict:
            self.central_focus = None

        return success

    def stop_stage(self):
        """Stop stage."""

        self.ask_stage_for_position = True

        for stage, axes in self.stages_list:
            stage.stop()

        self.central_focus = self.get_stage_position().get("f_pos", self.central_focus)

    def get_stage_position(self):
        """Get stage position.

        Returns
        -------
        stage_position : dict
            Dictionary of stage positions.
        """
        if self.ask_stage_for_position:
            # self.ret_pos_dict = {}
            for stage, axes in self.stages_list:
                temp_pos = stage.report_position()
                self.ret_pos_dict.update(temp_pos)
            self.ask_stage_for_position = False
        return self.ret_pos_dict

    def move_remote_focus(self, offset=None):
        """Move remote focus.

        Parameters
        ----------
        offset : float, optional
            Offset, by default None
        """
        readout_time = self.get_readout_time()
        exposure_times, sweep_times = self.calculate_exposure_sweep_times(readout_time)
        self.remote_focus_device.move(exposure_times, sweep_times, offset)

    def update_stage_limits(self, limits_flag=True):
        """Update stage limits.

        Parameters
        ----------
        limits_flag : bool, optional
            Limits flag, by default True
        """
        self.ask_stage_for_position = True
        for stage, _ in self.stages_list:
            stage.stage_limits = limits_flag

    def assemble_device_config_lists(self, device_name, device_name_dict):
        """Assemble device config lists.

        Parameters
        ----------
        device_name : str
            Device name.
        device_name_dict : dict
            Device name dictionary.

        Returns
        -------
        device_config_list : list
            Device configuration list.
        device_name_list : list
            Device name list.
        """
        device_config_list = []
        device_name_list = []

        try:
            devices = self.configuration["configuration"]["microscopes"][
                self.microscope_name
            ][device_name]
        except KeyError:
            # if no such device
            return [], [], False

        if type(devices) == ListProxy:
            i = 0
            for d in devices:
                device_config_list.append(d)
                if device_name in device_name_dict:
                    device_name_list.append(
                        build_ref_name("_", d[device_name_dict[device_name]])
                    )

                else:
                    device_name_list.append(build_ref_name("_", device_name, i))
                i += 1
            is_list = True

        else:
            device_config_list.append(devices)
            is_list = False

        return device_config_list, device_name_list, is_list

    def load_and_start_devices(
        self,
        device_name,
        is_list,
        device_name_list,
        device_ref_name,
        device_connection,
        name,
        i,
        plugin_devices,
    ):
        """Load and start devices.

        Parameters
        ----------
        device_name : str
            Device name.
        is_list : bool
            Is list.
        device_name_list : list
            Device name list.
        device_ref_name : str
            Device reference name.
        device_connection : str
            Device connection.
        name : str
            Name.
        i : int
            Index.
        plugin_devices : dict
            Plugin Devices
        """
        # Import start_device classes
        try:
            exec(
                f"start_{device_name}=importlib.import_module("
                f"'navigate.model.device_startup_functions').start_{device_name}"
            )
        except AttributeError:
            print(f"Could not import start_{device_name}")
            print(f"Could not load device {device_name}")

        # Start the devices
        if is_list:
            exec(
                f"self.{device_name}['{device_name_list[i]}'] = "
                f"start_{device_name}(name, device_connection, self.configuration, "
                f"i, self.is_synthetic, plugin_devices)"
            )
            if device_name in device_name_list[i]:
                self.info[device_name_list[i]] = device_ref_name
        else:
            exec(
                f"self.{device_name} = start_{device_name}(name, "
                f"device_connection, self.configuration, self.is_synthetic, plugin_devices)"
            )
            self.info[device_name] = device_ref_name

    def terminate(self):
        """Close hardware explicitly."""
        self.camera.close_camera()

        for k in self.galvo:
            self.galvo[k].turn_off()

        try:
            # Currently only for RemoteFocusEquipmentSolutions
            self.remote_focus_device.close_connection()
        except AttributeError:
            pass

        try:
            for stage, _ in self.stages_list:
                stage.close()
        except Exception as e:
            print(f"Stage delete failure: {e}")
        pass

    def run_command(self, command, *args):
        if command in self.commands:
            result = self.commands[command][1](*args)
            if result:
                device_name = self.commands[command][0]
                self.output_event_queue.put((device_name, result))
        else:
            print("unknown command in the Microscope:", command)



# serialport.py
from serial import Serial
from serial import SerialException
from serial import SerialTimeoutException
from serial import EIGHTBITS
from serial import PARITY_NONE
from serial import STOPBITS_ONE
from serial.tools import list_ports
import threading

import time


class TigerException(Exception):
    """
    Exception raised when error code from Tiger Console is received.

    Attributes:
        - command: error code received from Tiger Console
    """

    def __init__(self, code: str):
        """Initialize the TigerException class
        Parameters
        ----------
        code : str
            Error code received from Tiger Console
        """

        #: dict: Dictionary of error codes and their corresponding error messages
        self.error_codes = {
            ":N-1": "Unknown Command (Not Issued in TG-1000)",
            ":N-2": "Unrecognized Axis Parameter (valid axes are dependent on the "
            "controller)",
            ":N-3": "Missing parameters (command received requires an axis parameter "
            "such as x=1234)",
            ":N-4": "Parameter Out of Range",
            ":N-5": "Operation failed",
            ":N-6": "Undefined Error (command is incorrect, but the controller does "
            "not know exactly why.",
            ":N-7": "Invalid Card Address",
            ":N-21": "Serial Command halted by the HALT command",
        }
        #: str: Error code received from Tiger Console
        self.code = code

        #: str: Error message
        self.message = self.error_codes[code]

        # Gets the proper message based on error code received.
        super().__init__(self.message)
        # Sends message to base exception constructor for python purposes

    def __str__(self):
        """Overrides base Exception string to be displayed
        in traceback"""
        return f"{self.code} -> {self.message}"


class TigerController:
    """Tiger Controller class"""

    def __init__(self, com_port: str, baud_rate: int, verbose: bool = False):
        """Initialize the Tiger Controller class


        Parameters
        ----------
        com_port : str
            COM port of the Tiger Controller
        baud_rate : int
            Baud rate of the Tiger Controller
        verbose : bool
            If True, will print out messages to the console

        """
        #: Serial: Serial port object
        self.serial_port = Serial()
        #: str: COM port of the Tiger Controller
        self.com_port = com_port
        #: int: Baud rate of the Tiger Controller
        self.baud_rate = baud_rate
        #: bool: If True, will print out messages to the console
        self.verbose = verbose
        #: list[str]: Default axes sequence of the Tiger Controller
        self.default_axes_sequence = None
        #: list[float]: Maximum speeds of the Tiger Controller
        self._max_speeds = None
        #: threading.Event: Event to indicate if it is safe to write to the serial port
        self.safe_to_write = threading.Event()
        #: threading.Event.set(): Set the safe_to_write event
        self.safe_to_write.set()
        #: float: Last time a command was sent to the Tiger Controller
        self._last_cmd_send_time = time.perf_counter()

    @staticmethod
    def scan_ports() -> list[str]:
        """Scans for available COM ports

        Returns a sorted list of COM ports.

        Returns
        -------
        list[str]
            Sorted list of COM ports
        """
        com_ports = [port.device for port in list_ports.comports()]
        com_ports.sort(key=lambda value: int(value[3:]))
        return com_ports

    def connect_to_serial(
        self,
        rx_size: int = 12800,
        tx_size: int = 12800,
        read_timeout: int = 1,
        write_timeout: int = 1,
    ) -> None:
        """
        Connect to the serial port.

        Parameters
        ----------
        rx_size : int
            Size of the rx buffer
        tx_size : int
            Size of the tx buffer
        read_timeout : int
            Read timeout in seconds
        write_timeout : int
            Write timeout in seconds
        """
        self.serial_port.port = self.com_port
        self.serial_port.baudrate = self.baud_rate
        self.serial_port.parity = PARITY_NONE
        self.serial_port.bytesize = EIGHTBITS
        self.serial_port.stopbits = STOPBITS_ONE
        self.serial_port.xonoff = False
        self.serial_port.rtscts = False
        self.serial_port.dsrdtr = False
        self.serial_port.write_timeout = write_timeout
        self.serial_port.timeout = read_timeout

        # set the size of the rx and tx buffers before calling open
        self.serial_port.set_buffer_size(rx_size, tx_size)

        # try to open the serial port
        try:
            self.serial_port.open()
        except SerialException:
            self.report_to_console(
                f"SerialException: can't connect to {self.com_port} at "
                f"{self.baud_rate}!"
            )

        if self.is_open():
            # clear the rx and tx buffers
            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()
            # report connection status to user
            self.report_to_console("Connected to the serial port.")
            self.report_to_console(
                f"Serial port = {self.com_port} :: Baud rate = {self.baud_rate}"
            )

            #: list[str]: Default axes sequence of the Tiger Controller
            self.default_axes_sequence = self.get_default_motor_axis_sequence()

    def get_default_motor_axis_sequence(self) -> None:
        """Get the default motor axis sequence from the ASI device

        Returns
        -------
        list[str]
            Default motor axis sequence
        """
        self.send_command("BU X")
        response = self.read_response()
        lines = response.split("\r")
        for line in lines:
            if line.startswith("Motor Axes:"):
                default_axes_sequence = line[line.index(":") + 2 :].split(" ")[:-2]
                self.report_to_console(
                    "Get the default axes sequence from the ASI device " "successfully!"
                )
                break

        return default_axes_sequence

    ##### TODO: Modify these to accept dictionaries and send a
    #           single command for all axes
    def set_feedback_alignment(self, axis, aa):
        """Set the stage feedback alignment.

        Parameters
        ----------
        axis : str
            Stage axis
        """
        self.send_command(f"AA {axis}={aa}\r")
        self.read_response()
        self.send_command(f"AZ {axis}\r")
        self.read_response()

    def set_backlash(self, axis, val):
        """Enable/disable stage backlash correction.

        Parameters
        ----------
        axis : str
            Stage axis
        val : float
            Distance of anti-backlash motion [mm]
        """
        self.send_command(f"B {axis}={val:.7f}\r")
        self.read_response()

    def set_finishing_accuracy(self, axis, ac):
        """Set the stage finishing accuracy.

        Parameters
        ----------
        axis : str
            Stage axis
        ac : float
            Position error [mm]
        """
        self.send_command(f"PC {axis}={ac:.7f}\r")
        self.read_response()

    def set_error(self, axis, ac):
        """
        Set the stage drift error

        Parameters
        ----------
        axis : str
            Stage axis
        ac : float
            Position error [mm]
        """
        self.send_command(f"E {axis}={ac:.7f}\r")
        self.read_response()

    ##### END TODO #####

    def disconnect_from_serial(self) -> None:
        """Disconnect from the serial port if it's open."""
        if self.is_open():
            self.serial_port.close()
            self.report_to_console("Disconnected from the serial port.")

    def is_open(self) -> bool:
        """Returns True if the serial port exists and is open.

        Returns
        -------
        bool
            True if the serial port exists and is open
        """
        # short circuits if serial port is None
        return self.serial_port and self.serial_port.is_open

    def report_to_console(self, message: str) -> None:
        """Print message to the output device, usually the console.

        Parameters
        ----------
        message : str
            Message to print to the output device
        """
        # useful if we want to output data to something other than the console
        # (ui element etc)
        if self.verbose:
            print(message)

    def send_command(self, cmd: bytes) -> None:
        """Send a serial command to the device.

        Parameters
        ----------
        cmd : bytes
            Serial command to send to the device
        """
        # always reset the buffers before a new command is sent
        self.safe_to_write.wait()
        self.safe_to_write.clear()
        self.serial_port.read_all()
        self.serial_port.reset_input_buffer()
        self.serial_port.reset_output_buffer()

        # send the serial command to the controller
        self.report_to_console(cmd)
        command = bytes(f"{cmd}\r", encoding="ascii")
        try:
            self.serial_port.write(command)
        except SerialTimeoutException as e:
            print(f"Tiger Controller -- SerialTimeoutException: {e}")
            pass

    def read_response(self) -> str:
        """Read a line from the serial response.

        Returns
        -------
        str
            Response from the serial port
        """
        response = self.serial_port.readline()
        self.safe_to_write.set()

        response = response.decode(encoding="ascii")
        # Remove leading and trailing empty spaces
        self.report_to_console(f"Received Response: {response.strip()}")
        if response.startswith(":N"):
            raise TigerException(response)

        return response  # in case we want to read the response

    def moverel(self, x: int = 0, y: int = 0, z: int = 0) -> None:
        """Move the stage with a relative move on multiple axes.

        Parameters
        ----------
        x : int
            Relative move on the x-axis
        y : int
            Relative move on the y-axis
        z : int
            Relative move on the z-axis
        """
        self.send_command(f"MOVREL X={x} Y={y} Z={z}\r")
        self.read_response()

    def moverel_axis(self, axis: str, distance: float) -> None:
        """Move the stage with a relative move on one axis

        Parameters
        ----------
        axis : str
            Stage axis
        distance : float
            Relative move distance
        """
        self.send_command(f"MOVREL {axis}={round(distance, 6)}\r")
        self.read_response()

    def move(self, pos_dict) -> None:
        """Move the stage with an absolute move on multiple axes

        Parameters
        ----------
        pos_dict : dict
            Dictionary of the form {axis: position}
        """
        pos_str = " ".join(
            [f"{axis}={round(pos, 6)}" for axis, pos in pos_dict.items()]
        )
        self.send_command(f"MOVE {pos_str}\r")
        self.read_response()

    def move_axis(self, axis: str, distance: float) -> None:
        """Move the stage with an absolute move on one axis

        Parameters
        ----------
        axis : str
            Stage axis
        distance : float
            Absolute move distance
        """
        self.send_command(f"MOVE {axis}={round(distance, 6)}\r")
        self.read_response()

    def set_max_speed(self, axis: str, speed: float) -> None:
        """Set the speed on a specific axis. Speed is in mm/s.

        Parameters
        ----------
        axis : str
            Stage axis
        speed : float
            Speed in mm/s
        """
        self.send_command(f"SPEED {axis}={speed}\r")
        self.read_response()

    def get_axis_position(self, axis: str) -> int:
        """Return the position of the stage in ASI units (tenths of microns).

        Parameters
        ----------
        axis : str
            Stage axis
        Returns
        -------
        int
            Position of the stage in ASI units
        """
        self.send_command(f"WHERE {axis}\r")
        response = self.read_response()
        # try:
        pos = float(response.split(" ")[1])
        # except:
        #     pos = float('Inf')
        return pos

    def get_axis_position_um(self, axis: str) -> float:
        """Return the position of the stage in microns.

        Parameters
        ----------
        axis : str
            Stage axis
        Returns
        -------
        float
            Position of the stage in microns
        """
        self.send_command(f"WHERE {axis}\r")
        response = self.read_response()
        return float(response.split(" ")[1]) / 10.0

    def get_position(self, axes) -> dict:
        """Return current stage position in ASI units.

        If default axes sequence has gotten from the ASI device,
        then it will ask the device all the position in one command,
        else it will ask each axis position one by one.

        WATCH OUT! This will return the positions in the order
        of the underlying hardware no matter what order the axes
        are passed in.

        See https://asiimaging.com/docs/products/serial_commands#commandwhere_w

        Parameters
        ----------
        axes : list[str]
            List of axes to query

        Returns
        -------
        dictionary:
             {axis: position}
        """

        if self.default_axes_sequence:
            cmd = f"WHERE {' '.join(axes)}\r"
            self.send_command(cmd)
            response = self.read_response()

            # return response.split(" ")[1:-1]
            pos = response.split(" ")
            axes_seq = list(
                filter(
                    lambda axis: axis if axis in axes else False,
                    self.default_axes_sequence,
                )
            )
            # return {axis: float(pos[1 + i]) for i, axis in enumerate(axes_seq)}
            axis_dict = {}
            for i, axis in enumerate(axes_seq):
                try:
                    axis_dict[axis] = float(pos[1 + i])
                except (ValueError, IndexError):
                    # Report position failed. Don't crash, we can try again.
                    pass
            return axis_dict
        else:
            result = {}
            for axis in axes:
                result[axis] = self.get_axis_position(axis)
            return result

    # Utility Functions

    def is_axis_busy(self, axis: str) -> bool:
        """Returns True if the axis is busy.

        Parameters
        ----------
        axis : str
            Stage axis
        Returns
        -------
        bool
            True if the axis is busy
        """
        self.send_command(f"RS {axis}?\r")
        res = self.read_response()
        return "B" in res

    def is_device_busy(self) -> bool:
        """Returns True if any axis is busy.

        Returns
        -------
        bool
            True if any axis is busy
        """
        self.send_command("/")
        res = self.read_response()
        return "B" in res

    def wait_for_device(self, timeout: float = 1.75) -> None:
        """Waits for the all motors to stop moving.

        timeout : float
            Timeout in seconds. Default is 1.75 seconds.
        """
        if self.verbose:
            print("Waiting for device...")
        busy = self.is_device_busy()
        waiting_time = 0.0

        while busy:
            waiting_time += 0.001
            if waiting_time >= timeout:
                break
            time.sleep(0.001)
            busy = self.is_device_busy()

        if self.verbose:
            print(f"Waited {waiting_time:.2f} s")

    def stop(self):
        """Stop all stage movement immediately"""

        self.send_command("HALT")
        self.read_response()
        if self.verbose:
            print("ASI Stages stopped successfully")

    def set_speed(self, speed_dict):
        """Set speed

        Parameters
        ----------
        speed_dict : dict
            Dictionary of the form {axis: speed}
        """
        axes = " ".join([f"{x}={round(v, 6)}" for x, v in speed_dict.items()])
        self.send_command(f"S {axes}")
        self.read_response()

    def set_speed_as_percent_max(self, pct):
        """Set speed as a percentage of the maximum speed

        Parameters
        ----------
        pct : float
            Percentage of the maximum speed
        """
        if self.default_axes_sequence is None:
            raise TigerException(
                "Unable to query system for axis sequence. Cannot set speed."
            )
        if self._max_speeds is None:
            # First, set the speed crazy high
            self.send_command(
                f"SPEED {' '.join([f'{ax}=1000' for ax in self.default_axes_sequence])}\r"  # noqa
            )
            self.read_response()

            # Next query the maximum speed
            self.send_command(
                f"SPEED {' '.join([f'{ax}?' for ax in self.default_axes_sequence])}\r"
            )
            res = self.read_response()
            self._max_speeds = [float(x.split("=")[1]) for x in res.split()[1:]]

        # Now set to pct
        self.send_command(
            f"SPEED {' '.join([f'{ax}={pct*speed:.7f}' for ax, speed in zip(self.default_axes_sequence, self._max_speeds)])}\r"  # noqa
        )
        self.read_response()

    def get_speed(self, axis: str):
        """Get speed

        Parameters
        ----------
        axis : str
            Stage axis
        """
        self.send_command(f"SPEED {axis}?")
        response = self.read_response()
        return float(response.split("=")[1])

    def get_encoder_counts_per_mm(self, axis: str):
        """Get encoder counts pre mm of axis

        Parameters
        ----------
        axis : str
            Stage axis

        Returns
        -------
        float
            Encoder counts per mm of axis
        """

        self.send_command(f"CNTS {axis}?")
        response = self.read_response()
        return float(response.split("=")[1].split()[0])

    def scanr(
        self,
        start_position_mm: float,
        end_position_mm: float,
        enc_divide: float = 0,
        axis: str = "X",
    ):
        """Set scanr operation mode.

        Parameters
        ----------
        start_position_mm : float
            Start position in mm
        end_position_mm : float
            End position in mm
        enc_divide : float
            Encoder divide
        axis : str
            Stage axis
        """
        enc_divide_mm = self.get_encoder_counts_per_mm(axis)
        if enc_divide == 0:
            enc_divide = enc_divide_mm
        else:
            enc_divide = int(enc_divide * enc_divide_mm)

        command = (
            f"SCANR "
            f"X={round(start_position_mm, 6)} "
            f"Y={round(end_position_mm, 6)} "
            f"Z={round(enc_divide)}"
        )

        self.send_command(command)
        self.read_response()

    def scanv(
        self,
        start_position_mm: float,
        end_position_mm: float,
        number_of_lines: float,
        overshoot: float = 1.0,
        axis: str = "X",
    ):
        """Set scanv operation mode.

        Parameters
        ----------
        start_position_mm : float
            Start position in mm
        end_position_mm : float
            End position in mm
        number_of_lines : float
            Number of lines
        overshoot : float
            Overshoot
        axis : str
            Stage axis
        """
        command = (
            f"SCANV "
            f"X={round(start_position_mm, 6)} "
            f"Y={round(end_position_mm, 6)} "
            f"Z={round(number_of_lines, 6)}"
            f"F={round(overshoot, 6)}"
        )

        self.send_command(command)
        self.read_response()

    def start_scan(self, axis: str, is_single_axis_scan: bool = True):
        """
        Start scan

        Parameters
        ----------
        axis : str
            Stage axis
        is_single_axis_scan : bool
            If True, will only scan on one axis
        """
        self.send_command("SCAN")
        self.read_response()

    def stop_scan(self):
        """Stop scan."""
        self.send_command("SCAN P")
        self.read_response()

    def is_moving(self):
        """Check to see if the stage is moving.

        Sends the command / which is equivalent to STATUS

        Gets response:
        N - there are no motors running from a serial command
        B - there is a motor running from a serial command

        Returns
        -------
        response: Bool
            True if any axis is moving. False if not.
        """

        # Calculate duration of time since last command.
        time_since_last_cmd = time.perf_counter() - self._last_cmd_send_time

        # Wait 50 milliseconds before pinging controller again.
        sleep_time = 0.050 - time_since_last_cmd
        if sleep_time > 0:
            time.sleep(sleep_time)

        self.send_command("/")
        response = self.read_response().rstrip().rstrip("\r\n")
        if response == "ACK":
            self.send_command("/")
            response = self.read_response().rstrip().rstrip("\r\n")
        if response == "B":
            return True
        elif response == "N":
            return False
        else:
            print("WARNING: WAIT UNTIL DONE RECEIVED NO RESPONSE")
            return False

    # Basic Serial Commands for Filter Wheels

    def send_filter_wheel_command(self, cmd) -> None:
        """Send a serial command to the filter wheel.

        Parameters
        ----------
        cmd : str
            Serial command to send to the filter wheel
        """

        # send the serial command to the controller
        self.send_command(f"{cmd}\n")
        # print(f"Sent Command: {command.decode(encoding='ascii')}")

    def select_filter_wheel(self, filter_wheel_number=0):
        """
        Select the filter wheel, e.g., 0, 1...

        Sets the current filter wheel for subsequent commands. Prompt shows currently
        selected wheel, e.g., 0> is result of FW 0 command. If the selected wheel is
        HOMED and ready to go, the FW command returns the selected wheel as normal. If
        the wheel is not ready for any reason, the response ERR is returned. Example:

        0> FW 1 1 Normal – switch to FW 1
        1> FW 0 ERR FW 0 not ready
        0> Although FW 0 not ready – can still change FW 0 parameters.
        """
        assert filter_wheel_number in range(2)
        self.send_filter_wheel_command(f"FW {filter_wheel_number}")
        self.read_response()

    def move_filter_wheel(self, filter_wheel_position=0):
        """
        Move to filter position n , where n is a valid filter position.
        """
        assert filter_wheel_position in range(8)
        self.send_filter_wheel_command(f"MP {filter_wheel_position}")
        self.read_response()

    def move_filter_wheel_to_home(self):
        """
        Causes current wheel to seek its home position.
        """
        self.send_filter_wheel_command("HO")
        self.read_response()

    def change_filter_wheel_speed(self, speed=0):
        """
        Selects a consistent set of preset acceleration and speed parameters.
        Supported in version 2.4 and later.

        0	Default - directly set and saved AU, AD, and VR parameters are used.
        1	Slowest and smoothest switching speed.
        2 to 8	Intermediate switching speeds.
        9	Fastest and but least reliable switching speed.
        """
        self.send_filter_wheel_command(f"SV {speed}")
        self.read_response()

    def halt_filter_wheel(self):
        """
        Halt filter wheel
        """
        self.send_filter_wheel_command("HA")
        self.read_response()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ZoomBase:
    """ZoomBase parent class."""

    def __init__(self, microscope_name, device_controller, configuration):
        """Initialize the parent zoom class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_controller : object
            Hardware device to connect to
        configuration : multiprocesing.managers.DictProxy
            Global configuration of the microscope
        """

        #: dict: Configuration dictionary for the device.
        self.configuration = configuration["configuration"]["microscopes"][
            microscope_name
        ]["zoom"]

        #: dict: Zoom dictionary
        self.zoomdict = self.configuration["position"]
        self.build_stage_dict()

        #: float: the desired zoom setting
        self.zoomvalue = None

    def build_stage_dict(self):
        """
        Construct a dictionary of stage offsets in between different zoom values.

        This assumes self.configuration["stage_positions"] is a dictionary of ideal
        focus positions for the same object taken in the same solvent at
        each magnification. e.g.,

        stage_positions:
            BABB:
                f:
                    0.63x: 0
                    1x: 70775
                    2x: 72455
                    3x: 72710
                    4x: 72795
                    5x: 72850
                    6x: 72880

        The resulting dictionary is keyed first on solvent (refractive index),
        followed by the stage axis to offset, followed by the current magnification
        and then the target magnification.
        """
        try:
            stage_positions = self.configuration["stage_positions"]
        except KeyError:
            self.stage_offsets = None
            return

        #: dict: Dictionary of stage offsets in between different zoom values.
        self.stage_offsets = {}
        for solvent, axes in stage_positions.items():
            self.stage_offsets[solvent] = {}
            for axis, mags in axes.items():
                self.stage_offsets[solvent][axis] = {}
                for mag_curr, focus_curr in mags.items():
                    self.stage_offsets[solvent][axis][mag_curr] = {}
                    for mag_target, focus_target in mags.items():
                        self.stage_offsets[solvent][axis][mag_curr][mag_target] = (
                            focus_target - focus_curr
                        )

    def set_zoom(self, zoom, wait_until_done=False):
        """Change the microscope zoom.

        Confirms tha the zoom position is available in the zoomdict

        Parameters
        ----------
        zoom : dict
            Zoom dictionary
        wait_until_done : bool
            Delay parameter.

        Changes zoom after checking that the commanded value exists
        """
        if zoom in self.zoomdict:
            self.zoomvalue = zoom
        else:
            logger.error(f"Zoom designation, {zoom}, not in the configuration")
            raise ValueError("Zoom designation not in the configuration")
        logger.debug(f"Changed Zoom to {zoom}")
        logger.debug(f"Zoom position: {self.read_position()}")

    def move(self, position=0, wait_until_done=False):
        """Move the Zoom Servo

        Parameters
        ----------
        position : int
            Location to move to.
        wait_until_done : bool
            Delay parameter
        """
        logger.debug(f"Changing Zoom to {position}")
        pass

    def read_position(self):
        """Read the position of the Zoom Servo

        Returns
        -------
        cur_position : int
            Current position of Zoom
        """
        cur_position = None
        logger.debug(f"Zoom position: {cur_position}")
        return cur_position



import time
import cv2
import numpy as np

from pyvcam import pvc
from pyvcam.camera import Camera

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()
    cam.readout_port = 0

    cam.prog_scan_mode = 1  # Scan mode options: {'Auto': 0, 'Line Delay': 1, 'Scan Width': 2}
    cam.prog_scan_dir = 0  # Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}


    def calculate_ASLMparameters(desired_exposuretime):
        """
        calculate the parameters for an ASLM acquisition
        :param desired_exposuretime: the exposure time that is desired for the whole acquisition
        :return: set the important parameters for ASLM acquisitions
        """
        linedelay = 0.00375 #10.16us
        nbrows = 3200
        ASLM_scanWidth=70

        ASLM_lineExposure = int(np.ceil(desired_exposuretime / (1 + nbrows / ASLM_scanWidth)))
        ASLM_line_delay = int(np.ceil((desired_exposuretime - ASLM_lineExposure) / (nbrows * linedelay))) - 1
        ASLM_acquisition_time = (ASLM_line_delay + 1) * nbrows * linedelay + ASLM_lineExposure + (ASLM_line_delay + 1) * linedelay

        print(
            "ASLM parameters are: {} exposure time, and {} line delay factor, {} total acquisition time for {} scan width".format(
                ASLM_lineExposure, ASLM_line_delay, ASLM_acquisition_time, ASLM_scanWidth))
        return ASLM_lineExposure, ASLM_line_delay

    exp, linedelay = calculate_ASLMparameters(200)
    cam.prog_scan_line_delay = linedelay


    cam.start_live(exp_time=exp)
    cnt = 0
    tot = 0
    t1 = time.time()
    start = time.time()
    width = 800
    height = int(cam.sensor_size[1] * width / cam.sensor_size[0])
    dim = (width, height)
    fps = 0

    while True:
        frame, fps, frame_count = cam.poll_frame()
        frame['pixel_data'] = cv2.resize(frame['pixel_data'], dim, interpolation = cv2.INTER_AREA)
        cv2.imshow('Live Mode', frame['pixel_data'])

        low = np.amin(frame['pixel_data'])
        high = np.amax(frame['pixel_data'])
        average = np.average(frame['pixel_data'])

        if cnt == 10:
                t1 = time.time() - t1
                fps = 10/t1
                t1 = time.time()
                cnt = 0
        if cv2.waitKey(10) == 27:
            break
        #print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Rate: {:.1f}\n'.format(low, high, average, fps))
        cnt += 1
        tot += 1

    cam.finish()
    cam.close()
    pvc.uninit_pvcam()

    print('Total frames: {}\nAverage fps: {}\n'.format(tot, (tot/(time.time()-start))))
if __name__ == "__main__":
    main()



"""
Sub-controllers for navigate.
"""

from .stage_controller import StageController  # noqa
from .acquire_bar_controller import AcquireBarController  # noqa
from .channels_tab_controller import ChannelsTabController  # noqa
from .camera_view_controller import CameraViewController  # noqa
from .camera_setting_controller import CameraSettingController  # noqa
from .waveform_tab_controller import WaveformTabController  # noqa
from .waveform_popup_controller import WaveformPopupController  # noqa
from .autofocus_popup_controller import AutofocusPopupController  # noqa
from .features_popup_controller import FeaturePopupController  # noqa
from .feature_advanced_setting_controller import (
    FeatureAdvancedSettingController,  # noqa
)
from .keystroke_controller import KeystrokeController  # noqa
from .multi_position_controller import MultiPositionController  # noqa
from .ilastik_popup_controller import IlastikPopupController  # noqa
from .camera_map_setting_popup_controller import CameraMapSettingPopupController  # noqa
from .microscope_popup_controller import MicroscopePopupController  # noqa
from .adaptiveoptics_popup_controller import AdaptiveOpticsPopupController  # noqa
# from .uninstall_plugin_controller import UninstallPluginController  # noqa
from .plugins_controller import PluginsController, UninstallPluginController  # noqa
from .menu_controller import MenuController  # noqa


# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import threading
import logging
import multiprocessing as mp
import time
import os

# Third Party Imports

# Local Imports
from navigate.model.concurrency.concurrency_tools import SharedNDArray
from navigate.model.features.autofocus import Autofocus
from navigate.model.features.cva_conpro import ConstantVelocityAcquisition
from navigate.model.features.adaptive_optics import TonyWilson
from navigate.model.features.image_writer import ImageWriter
from navigate.model.features.auto_tile_scan import CalculateFocusRange  # noqa
from navigate.model.features.common_features import (
    ChangeResolution,
    Snap,
    ZStackAcquisition,
    FindTissueSimple2D,
    PrepareNextChannel,
    LoopByCount,
    StackPause,
    MoveToNextPositionInMultiPositionTable,
    WaitToContinue,
)
from navigate.model.features.remove_empty_tiles import (
    DetectTissueInStackAndRecord,
    RemoveEmptyPositions,
)
from navigate.model.features.feature_container import load_features
from navigate.model.features.restful_features import IlastikSegmentation
from navigate.model.features.volume_search import VolumeSearch
from navigate.model.features.feature_related_functions import (
    convert_str_to_feature_list,
    convert_feature_list_to_str,
    SharedList,
    load_dynamic_parameter_functions,
)
from navigate.log_files.log_functions import log_setup
from navigate.tools.common_dict_tools import update_stage_dict
from navigate.tools.common_functions import load_module_from_file
from navigate.tools.file_functions import load_yaml_file, save_yaml_file
from navigate.model.device_startup_functions import load_devices
from navigate.model.microscope import Microscope
from navigate.config.config import get_navigate_path
from navigate.model.plugins_model import PluginsModel


# Logger Setup
p = __name__.split(".")[1]


class Model:
    """Navigate Model Class

    Model for Model-View-Controller Software Architecture."""

    def __init__(self, args, configuration=None, event_queue=None):
        """Initialize the Model.

        Parameters
        ----------
        args : argparse.Namespace
            Command line arguments.
        configuration : dict
            Configuration dictionary.
        event_queue : multiprocessing.Queue
            Queue for events.
        """

        log_setup("model_logging.yml")
        #: object: Logger object.
        self.logger = logging.getLogger(p)

        # Loads the YAML file for all of the microscope parameters
        #: dict: Configuration dictionary.
        self.configuration = configuration

        plugins = PluginsModel()
        # load plugin feature and devices
        plugin_devices, plugin_acquisition_modes = plugins.load_plugins()
        devices_dict = load_devices(
            configuration, args.synthetic_hardware, plugin_devices
        )
        devices_dict["__plugins__"] = plugin_devices

        #: dict: Dictionary of plugin acquisition modes
        self.plugin_acquisition_modes = plugin_acquisition_modes

        #: dict: Dictionary of virtual microscopes.
        self.virtual_microscopes = {}
        #: dict: Dictionary of physical microscopes.
        self.microscopes = {}
        for microscope_name in configuration["configuration"]["microscopes"].keys():
            self.microscopes[microscope_name] = Microscope(
                microscope_name, configuration, devices_dict, args.synthetic_hardware
            )
            self.microscopes[microscope_name].output_event_queue = event_queue
        # register device commands if there is any.

        #: str: Name of the active microscope.
        self.active_microscope = None
        #: str: Name of the active microscope.
        self.active_microscope_name = None
        self.get_active_microscope()

        # Acquisition Housekeeping
        #: str: Imaging mode.
        self.imaging_mode = None
        #: int: Number of images acquired.
        self.image_count = 0
        #: int: Number of acquisitions.
        self.acquisition_count = 0
        #: int: Total number of acquisitions.
        self.total_acquisition_count = None
        #: int: Total number of images.
        self.total_image_count = None
        #: float: Current exposure time in milliseconds
        self.current_exposure_time = 0  # milliseconds
        #: float: Pre-exposure time in milliseconds
        self.pre_exposure_time = 0  # milliseconds
        #: int: Number of timeouts before aborting acquisition.
        self.camera_wait_iterations = 20  # Thread waits this * 500 ms before it ends
        #: float: Time before acquisition.
        self.start_time = None
        #: object: Data buffer.
        self.data_buffer = None
        #: int: Number of active pixels in the x-dimension.
        self.img_width = int(
            self.configuration["experiment"]["CameraParameters"]["img_x_pixels"]
        )
        #: int: Number of active pixels in the y-dimension.
        self.img_height = int(
            self.configuration["experiment"]["CameraParameters"]["img_y_pixels"]
        )
        #: str: Binning mode.
        self.binning = "1x1"
        #: array: stage positions.
        self.data_buffer_positions = None
        #: array: saving flags for a frame
        self.data_buffer_saving_flags = None
        #: bool: Is the model acquiring?
        self.is_acquiring = False

        # Autofocusing
        #: float: Current focus position.
        self.f_position = None
        #: float: Autofocus maximum entropy.
        self.max_entropy = None
        #: float: Autofocus maximum entropy position.
        self.focus_pos = None

        # Threads
        #: threading.Thread: Signal thread.
        self.signal_thread = None
        #: threading.Thread: Data thread.
        self.data_thread = None

        # show image function/pipe handler
        #: multiprocessing.connection.Connection: Show image pipe.
        self.show_img_pipe = None

        # Plot Pipe handler
        #: multiprocessing.connection.Connection: Plot pipe.
        self.plot_pipe = None

        # waveform queue
        #: multiprocessing.Queue: Waveform queue.
        self.event_queue = event_queue

        # frame signal id
        #: int: Frame ID.
        self.frame_id = 0

        # flags
        #: bool: Autofocus on?
        self.autofocus_on = False  # autofocus
        #: bool: Is the model live?
        self.is_live = False  # need to clear up data buffer after acquisition
        #: bool: Is the model saving the data?
        self.is_save = False  # save data
        #: bool: Stop signal and data threads?
        self.stop_acquisition = False  # stop signal and data threads
        #: bool: Stop signal thread?
        self.stop_send_signal = False  # stop signal thread
        #: event: Pause data event.
        self.pause_data_event = threading.Event()
        #: threading.Lock: Pause data ready lock.
        self.pause_data_ready_lock = threading.Lock()
        #: bool: Ask to pause data thread?
        self.ask_to_pause_data_thread = False

        # data buffer for image frames
        #: int: Number of frames in the data buffer.
        self.number_of_frames = self.configuration["experiment"]["CameraParameters"][
            "databuffer_size"
        ]
        self.update_data_buffer(self.img_width, self.img_height)

        # Image Writer/Save functionality
        #: ImageWriter: Image writer.
        self.image_writer = None

        # feature list
        #: list: add on feature in customized mode
        self.addon_feature = None
        #: list: List of features.
        self.feature_list = []
        # automatically switch resolution
        self.feature_list.append(
            [
                {"name": ChangeResolution, "args": ("Mesoscale", "1x")},
                {"name": Snap},
            ]
        )
        # z stack acquisition
        self.feature_list.append([{"name": ZStackAcquisition}])
        # threshold and tile
        self.feature_list.append([{"name": FindTissueSimple2D}])
        # Ilastik segmentation
        self.feature_list.append([{"name": IlastikSegmentation}])
        # volume search
        self.feature_list.append(
            [
                {
                    "name": VolumeSearch,
                    "args": ("Nanoscale", "N/A", True, False, 0.1),
                }
            ]
        )
        self.feature_list.append(
            [
                (
                    (
                        {"name": PrepareNextChannel},
                        {
                            "name": LoopByCount,
                            "args": ("experiment.MicroscopeState.selected_channels",),
                        },
                    ),
                    {
                        "name": LoopByCount,
                        "args": ("experiment.MicroscopeState.timepoints",),
                    },
                )
            ]
        )

        self.feature_list.append(
            [
                # {"name": MoveToNextPositionInMultiPositionTable},
                # {"name": CalculateFocusRange},
                {"name": PrepareNextChannel},
                (
                    {"name": MoveToNextPositionInMultiPositionTable},
                    {"name": Autofocus},
                    {
                        "name": ZStackAcquisition,
                        "args": (
                            True,
                            True,
                        ),
                    },
                    {"name": WaitToContinue},
                    {
                        "name": LoopByCount,
                        "args": ("experiment.MicroscopeState.multiposition_count",),
                    },
                ),
            ]
        )

        records = SharedList([], "records")
        self.feature_list.append(
            [
                {"name": PrepareNextChannel},
                (
                    {"name": MoveToNextPositionInMultiPositionTable},
                    # {"name": CalculateFocusRange},
                    {
                        "name": DetectTissueInStackAndRecord,
                        "args": (
                            5,
                            0.75,
                            records,
                        ),
                    },
                    {
                        "name": LoopByCount,
                        "args": ("experiment.MicroscopeState.multiposition_count",),
                    },
                ),
                {"name": RemoveEmptyPositions, "args": (records,)},
            ]
        )

        self.acquisition_modes_feature_setting = {
            "single": [
                (
                    {"name": PrepareNextChannel},
                    {
                        "name": LoopByCount,
                        "args": ("experiment.MicroscopeState.selected_channels",),
                    },
                )
            ],
            "live": [
                (
                    {"name": PrepareNextChannel},
                    {
                        "name": LoopByCount,
                        "args": ("experiment.MicroscopeState.selected_channels",),
                    },
                )
            ],
            "z-stack": [
                (
                    {"name": ZStackAcquisition},
                    {"name": StackPause},
                    {
                        "name": LoopByCount,
                        "args": ("experiment.MicroscopeState.timepoints",),
                    },
                )
            ],
            "ConstantVelocityAcquisition": [{"name": ConstantVelocityAcquisition}],
            "customized": [],
        }
        # append plugin acquisition mode
        for mode in self.plugin_acquisition_modes:
            self.acquisition_modes_feature_setting[
                mode
            ] = self.plugin_acquisition_modes[mode].feature_list

        self.load_feature_records()

    def update_data_buffer(self, img_width=512, img_height=512):
        """Update the Data Buffer

        Parameters
        ----------
        img_width : int
            Number of active pixels in the x-dimension.
        img_height : int
            Number of active pixels in the y-dimension.
        """
        self.img_width = img_width
        self.img_height = img_height
        self.data_buffer = [
            SharedNDArray(shape=(img_height, img_width), dtype="uint16")
            for i in range(self.number_of_frames)
        ]
        self.data_buffer_positions = SharedNDArray(
            shape=(self.number_of_frames, 5), dtype=float
        )  # z-index, x, y, z, theta, f
        for microscope_name in self.microscopes:
            self.microscopes[microscope_name].update_data_buffer(
                self.configuration["experiment"]["CameraParameters"]["x_pixels"],
                self.configuration["experiment"]["CameraParameters"]["y_pixels"],
                self.data_buffer,
                self.number_of_frames,
            )

    def get_data_buffer(self, img_width=512, img_height=512):
        """Get the data buffer.

        If the number of active pixels in x and y changes, updates the data buffer and
        returns newly-sized buffer.

        Parameters
        ----------
        img_height : int
            Number of active pixels in the x-dimension.
        img_width : int
            Number of active pixels in the y-dimension.

        Returns
        -------
        data_buffer : SharedNDArray
            Shared memory object.
        """
        if (
            img_width != self.img_width
            or img_height != self.img_height
            or self.configuration["experiment"]["CameraParameters"]["binning"]
            != self.binning
        ):
            self.update_data_buffer(img_width, img_height)
        return self.data_buffer

    def create_pipe(self, pipe_name):
        """Create a data pipe.

        Creates a pair of connection objects connected by a pipe which by default is
        duplex (two-way)

        Parameters
        ----------
        pipe_name : str
            Name of pipe to create.

        Returns
        -------
        end1 : object
            Connection object.
        """
        self.release_pipe(pipe_name)
        end1, end2 = mp.Pipe()
        setattr(self, pipe_name, end2)
        return end1

    def release_pipe(self, pipe_name):
        """Close a data pipe.

        Parameters
        ----------
        pipe_name : str
            Name of pipe to close.
        """
        if hasattr(self, pipe_name):
            pipe = getattr(self, pipe_name)
            if pipe:
                pipe.close()
            delattr(self, pipe_name)

    def get_active_microscope(self):
        """Get the active microscope.

        Returns
        -------
        microscope : Microscope
            Active microscope.
        """

        self.active_microscope_name = self.configuration["experiment"][
            "MicroscopeState"
        ]["microscope_name"]
        self.active_microscope = self.microscopes[self.active_microscope_name]
        return self.active_microscope

    def get_offset_variance_maps(self):
        """Get the offset variance maps.

        Returns
        -------
        offset_variance_maps : dict
            Offset variance maps.
        """

        return self.active_microscope.camera.get_offset_variance_maps()

    def run_command(self, command, *args, **kwargs):
        """Receives commands from the controller.

        Parameters
        ----------
        command : str
            Type of command to run.
        *args : list
            List of arguments to pass to the command.
        **kwargs : dict
            Dictionary of keyword arguments to pass to the command.
        """
        logging.info(
            "Navigate Model - Received command from controller:", command, args
        )
        if not self.data_buffer:
            logging.debug("Navigate Model - Shared Memory Buffer Not Set Up.")
            return

        if command == "acquire":
            self.is_acquiring = True
            self.imaging_mode = self.configuration["experiment"]["MicroscopeState"][
                "image_mode"
            ]
            self.is_save = self.configuration["experiment"]["MicroscopeState"][
                "is_save"
            ]

            # If multiposition is selected, verify that it is not empty.
            if self.configuration["experiment"]["MicroscopeState"]["is_multiposition"]:
                if len(self.configuration["experiment"]["MultiPositions"]) == 0:
                    # Update the view and override the settings.
                    self.event_queue.put(("disable_multiposition", None))
                    self.configuration["experiment"]["MicroscopeState"][
                        "is_multiposition"
                    ] = False

            # Calculate waveforms, turn on lasers, etc.
            self.prepare_acquisition()

            # load features
            if self.imaging_mode == "customized":
                if self.addon_feature is None:
                    self.addon_feature = self.acquisition_modes_feature_setting[
                        "single"
                    ]
                self.signal_container, self.data_container = load_features(
                    self, self.addon_feature
                )
                self.data_buffer_saving_flags = [False] * self.number_of_frames
            else:
                self.signal_container, self.data_container = load_features(
                    self, self.acquisition_modes_feature_setting[self.imaging_mode]
                )
                self.data_buffer_saving_flags = None

            if self.imaging_mode == "live":
                self.signal_thread = threading.Thread(target=self.run_live_acquisition)
            else:
                self.signal_thread = threading.Thread(target=self.run_acquisition)

            self.signal_thread.name = f"{self.imaging_mode} signal"
            if self.is_save and self.imaging_mode != "live":
                saving_config = {}
                plugin_obj = self.plugin_acquisition_modes.get(self.imaging_mode, None)
                if plugin_obj and hasattr(plugin_obj, "update_saving_config"):
                    saving_config = getattr(plugin_obj, "update_saving_config")(self)
                self.image_writer = ImageWriter(
                    self,
                    saving_flags=self.data_buffer_saving_flags,
                    saving_config=saving_config,
                )
                self.data_thread = threading.Thread(
                    target=self.run_data_process,
                    kwargs={"data_func": self.image_writer.save_image},
                )
            else:
                self.is_save = False
                self.data_thread = threading.Thread(target=self.run_data_process)
            self.data_thread.name = f"{self.imaging_mode} Data"
            self.signal_thread.start()
            self.data_thread.start()
            for m in self.virtual_microscopes:
                image_writer = (
                    ImageWriter(
                        self,
                        self.virtual_microscopes[m].data_buffer,
                        m,
                        saving_flags=self.data_buffer_saving_flags,
                        saving_config=saving_config,
                    ).save_image
                    if self.is_save
                    else None
                )
                threading.Thread(
                    target=self.simplified_data_process,
                    args=(
                        self.virtual_microscopes[m],
                        getattr(self, f"{m}_show_img_pipe"),
                        image_writer,
                    ),
                ).start()

        elif command == "update_setting":
            """
            Called by the controller
            Passes the string 'resolution' and a dictionary
            consisting of the resolution_mode, the zoom, and the laser_info.
            e.g., self.resolution_info['waveform_constants'][self.resolution][self.mag]
            """
            reboot = False
            microscope_name = self.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ]
            if self.is_acquiring:
                # We called this while in the middle of an acquisition
                # stop live thread
                self.stop_send_signal = True
                self.signal_thread.join()
                if microscope_name != self.active_microscope_name:
                    self.pause_data_thread()
                    self.active_microscope.end_acquisition()
                    reboot = True
                self.active_microscope.current_channel = 0

            if args[0] == "resolution":
                self.change_resolution(
                    self.configuration["experiment"]["MicroscopeState"][
                        "microscope_name"
                    ]
                )

            if reboot:
                # prepare active microscope
                waveform_dict = self.active_microscope.prepare_acquisition()
                self.resume_data_thread()
            else:
                waveform_dict = self.active_microscope.calculate_all_waveform()

            self.event_queue.put(("waveform", waveform_dict))

            if self.is_acquiring:
                # prepare devices based on updated info
                # load features
                self.signal_container, self.data_container = load_features(
                    self, self.acquisition_modes_feature_setting[self.imaging_mode]
                )
                self.stop_send_signal = False
                self.signal_thread = threading.Thread(target=self.run_live_acquisition)
                self.signal_thread.name = "Waveform Popup Signal"
                self.signal_thread.start()

        elif command == "autofocus":
            """Autofocus Routine

            Parameters
            ----------
            Args[0]: device name
            Args[1]: device reference
            """
            if self.is_acquiring and self.imaging_mode == "live":
                if hasattr(self, "signal_container"):
                    self.signal_container.cleanup()
                if hasattr(self, "data_container"):
                    self.data_container.cleanup()
                self.signal_container, self.data_container = load_features(
                    self,
                    [{"name": Autofocus}],
                )
            elif not self.is_acquiring:
                self.is_acquiring = True
                autofocus = Autofocus(self, *args)
                autofocus.run()

        elif command == "flatten_mirror":
            self.update_mirror(coef=[], flatten=True)
        elif command == "zero_mirror":
            self.active_microscope.mirror.zero_flatness()
        elif command == "set_mirror":
            coef = list(
                self.configuration["experiment"]["MirrorParameters"]["modes"].values()
            )
            self.update_mirror(coef=coef)
        elif command == "save_wcs_file":
            self.active_microscope.mirror.save_wcs_file(path=args[0])
        elif command == "set_mirror_from_wcs":
            coef = self.active_microscope.mirror.set_from_wcs_file(path=args[0])
            self.update_mirror(coef=coef)
        elif command == "tony_wilson":
            tony_wilson = TonyWilson(self)
            tony_wilson.run(*args)

        elif command == "load_feature":
            """
            args[0]: int, args[0]-1 is the id of features
                   : 0 no features
                   : str, name of feature, case sensitive
            """
            if hasattr(self, "signal_container"):
                delattr(self, "signal_container")
                delattr(self, "data_container")

            if type(args[0]) == int:
                self.addon_feature = None
                if args[0] != 0:
                    if len(args) == 2:
                        self.feature_list[args[0] - 1] = convert_str_to_feature_list(
                            args[1]
                        )

                    self.addon_feature = self.feature_list[args[0] - 1]
                    load_dynamic_parameter_functions(
                        self.addon_feature,
                        f"{get_navigate_path()}/feature_lists/feature_parameter_setting",
                    )
                    self.signal_container, self.data_container = load_features(
                        self, self.addon_feature
                    )
            elif type(args[0]) == str:
                try:
                    if len(args) > 1:
                        self.addon_feature = [
                            {"name": globals()[args[0]], "args": (args[1],)}
                        ]
                        self.signal_container, self.data_container = load_features(
                            self, self.addon_feature
                        )
                    else:
                        self.addon_feature = [{"name": globals()[args[0]]}]
                        self.signal_container, self.data_container = load_features(
                            self, self.addon_feature
                        )
                except KeyError:
                    self.logger.debug(
                        f"run_command - load_feature - Unknown feature {args[0]}."
                    )
        elif command == "stage_limits":
            for microscope_name in self.microscopes:
                self.microscopes[microscope_name].update_stage_limits(args[0])
        elif command == "stop":
            """
            Called when user halts the acquisition
            """
            self.logger.info("Navigate Model - Stopping with stop command.")
            self.stop_acquisition = True

            if hasattr(self, "signal_container"):
                self.signal_container.end_flag = True
            if self.imaging_mode == "ConstantVelocityAcquisition":
                self.active_microscope.stages["z"].stop()
            if self.signal_thread:
                self.signal_thread.join()
            if self.data_thread:
                self.data_thread.join()

            self.end_acquisition()
            self.stop_stage()

        elif command == "terminate":
            self.terminate()

        # elif command == "change_camera":
        #     new_camera = list(self.active_microscope.cameras.values())[args[0]]
        #     print(f"Using new camera >> {new_camera.camera_controller._serial_number}")
        #     self.active_microscope.camera = new_camera

        elif command == "exit":
            for camera in self.active_microscope.cameras.values():
                camera.camera_controller.dev_close()
        else:
            self.active_microscope.run_command(command, args)

    # main function to update mirror/set experiment mode values
    def update_mirror(self, coef=[], flatten=False):
        """Update the mirror.

        Parameters
        ----------
        coef : list
            List of coefficients.
        flatten : bool
            Flatten the mirror?
        """
        if coef:
            self.active_microscope.mirror.display_modes(coef)
        elif flatten:
            self.active_microscope.mirror.flat()

        mirror_img = self.active_microscope.mirror.mirror_controller.get_wavefront_pix()

        self.event_queue.put(
            ("mirror_update", {"mirror_img": mirror_img, "coefs": coef})
        )

        # print(self.configuration['experiment']['MirrorParameters']['modes'])

    def move_stage(self, pos_dict, wait_until_done=False):
        """Moves the stages.

        Updates the stage dictionary, moves to the desired position, and reports
        the position.

        Parameters
        ----------
        pos_dict : dict
            Dictionary of stage positions.
        wait_until_done : bool
            Checks "on target state" after command and waits until done.

        Returns
        -------
        success : bool
            Was the move successful?
        """
        try:
            r = self.active_microscope.move_stage(pos_dict, wait_until_done)
        except Exception as e:
            self.logger.debug(f"*** stage move failed: {e}")
            return False
        return r

    def get_stage_position(self):
        """Get the position of the stage.

        Returns
        -------
        ret_pos_dict : dict
            Dictionary of stage positions.
        """
        return self.active_microscope.get_stage_position()

    def stop_stage(self):
        """Stop the stages."""
        self.active_microscope.stop_stage()
        ret_pos_dict = self.get_stage_position()
        update_stage_dict(self, ret_pos_dict)
        self.event_queue.put(("update_stage", ret_pos_dict))

    def end_acquisition(self):
        """End the acquisition.

        Sets the current channel to 0, clears the signal and data containers,
        disconnects buffer in live mode and closes the shutters."""
        self.is_acquiring = False

        self.active_microscope.end_acquisition()
        for microscope_name in self.virtual_microscopes:
            self.virtual_microscopes[microscope_name].end_acquisition()

        plugin_obj = self.plugin_acquisition_modes.get(self.imaging_mode, None)
        if plugin_obj and hasattr(plugin_obj, "end_acquisition_model"):
            getattr(plugin_obj, "end_acquisition_model")(self)

        if hasattr(self, "signal_container"):
            self.signal_container.cleanup()
            delattr(self, "signal_container")
        if hasattr(self, "data_container"):
            self.data_container.cleanup()
            delattr(self, "data_container")
        if self.image_writer is not None:
            self.image_writer.close()
        #: obj: Add on feature.
        self.addon_feature = None

    def run_data_process(self, num_of_frames=0, data_func=None):
        """Run the data process.

        This function is the structure of data thread.

        Parameters
        ----------
        num_of_frames : int
            Number of frames to acquire.
        data_func : object
            Function to run on the acquired data.
        """
        wait_num = self.camera_wait_iterations
        acquired_frame_num = 0

        # whether acquire specific number of frames.
        count_frame = num_of_frames > 0
        start_time = time.time()

        while not self.stop_acquisition:
            if self.ask_to_pause_data_thread:
                self.pause_data_ready_lock.release()
                self.pause_data_event.clear()
                self.pause_data_event.wait()
            frame_ids = self.active_microscope.camera.get_new_frame()
            self.logger.info(
                f"Navigate Model - Running data process, get frames {frame_ids}"
            )
            # if there is at least one frame available
            if not frame_ids:
                self.logger.info(f"Navigate Model - Waiting {wait_num}")
                wait_num -= 1
                if wait_num <= 0:
                    # Camera timeout, abort acquisition.
                    break
                continue

            acquired_frame_num += len(frame_ids)
            stop_time = time.time()
            frames_per_second = acquired_frame_num / (stop_time - start_time)
            self.event_queue.put(("framerate", frames_per_second))

            wait_num = self.camera_wait_iterations

            if hasattr(self, "data_container") and not self.data_container.end_flag:
                if self.data_container.is_closed:
                    self.logger.info("Navigate Model - Data container is closed.")
                    self.stop_acquisition = True
                    break

                self.data_container.run(frame_ids)

            # ImageWriter to save images
            if data_func:
                data_func(frame_ids)

            # show image
            self.logger.info(f"Navigate Model - Sent through pipe{frame_ids[0]}")
            self.show_img_pipe.send(frame_ids[-1])

            if count_frame and acquired_frame_num >= num_of_frames:
                self.logger.info("Navigate Model - Loop stop condition met.")
                self.stop_acquisition = True

        self.show_img_pipe.send("stop")
        self.logger.info("Navigate Model - Data thread stopped.")
        self.logger.info(
            f"Navigate Model - Received frames in total: {acquired_frame_num}"
        )

        # release the lock when data thread ends
        if self.pause_data_ready_lock.locked():
            self.pause_data_ready_lock.release()

        self.end_acquisition()  # Need this to turn off the lasers/close the shutters

    def pause_data_thread(self):
        """Pause the data thread.

        Function is called when user pauses the acquisition.
        """

        self.pause_data_ready_lock.acquire()
        self.ask_to_pause_data_thread = True
        self.pause_data_ready_lock.acquire()

    def resume_data_thread(self):
        """Resume the data thread.

        Function is called when user resumes the acquisition.
        """

        self.ask_to_pause_data_thread = False
        self.pause_data_event.set()
        if self.pause_data_ready_lock.locked():
            self.pause_data_ready_lock.release()

    def simplified_data_process(self, microscope, show_img_pipe, data_func=None):
        """Run the data process.

        Parameters
        ----------
        microscope : Microscope
            Microscope object.
        show_img_pipe : multiprocessing.connection.Connection
            Pipe for showing images.
        data_func : object
            Function to run on the acquired data.
        """

        wait_num = self.camera_wait_iterations
        acquired_frame_num = 0

        while not self.stop_acquisition:
            frame_ids = (
                microscope.camera.get_new_frame()
            )  # This is the 500 ms wait for Hamamatsu
            self.logger.info(
                f"Navigate Model - Running data process, get frames {frame_ids} from "
                f"{microscope.microscope_name}"
            )
            # if there is at least one frame available
            if not frame_ids:
                self.logger.info(f"Navigate Model - Waiting {wait_num}")
                wait_num -= 1
                if wait_num <= 0:
                    # Camera timeout, abort acquisition.
                    break
                continue

            wait_num = self.camera_wait_iterations

            # Leave it here for now to work with current ImageWriter workflow
            # Will move it feature container later
            if data_func:
                data_func(frame_ids)

            # show image
            self.logger.info(
                f"Navigate Model - Sent through pipe{frame_ids[0]} -- "
                f"{microscope.microscope_name}"
            )
            show_img_pipe.send(frame_ids[-1])
            acquired_frame_num += len(frame_ids)

        show_img_pipe.send("stop")
        self.logger.info("Navigate Model - Data thread stopped.")
        self.logger.info(
            f"Navigate Model - Received frames in total: {acquired_frame_num}"
        )

    def prepare_acquisition(self, turn_off_flags=True):
        """Prepare the acquisition.

        This function is called when user starts the acquisition.
        Sets flags. Calculates all of the waveforms. Sets the Camera Sensor Mode
        Initializes the data buffer and starts camera. Opens Shutters

        Parameters
        ----------
        turn_off_flags : bool
            Turn off the flags.
        """
        # turn off flags
        if turn_off_flags:
            self.stop_acquisition = False
            self.stop_send_signal = False
            self.autofocus_on = False
            self.is_live = False

        plugin_obj = self.plugin_acquisition_modes.get(self.imaging_mode, None)
        if plugin_obj and hasattr(plugin_obj, "prepare_acquisition_model"):
            getattr(plugin_obj, "prepare_acquisition_model")(self)

        for m in self.virtual_microscopes:
            self.virtual_microscopes[m].prepare_acquisition()

        # Confirm stage position and software are in agreement.
        self.stop_stage()

        # prepare active microscope
        waveform_dict = self.active_microscope.prepare_acquisition()
        self.event_queue.put(("waveform", waveform_dict))

        self.frame_id = 0

    def snap_image(self):
        """Acquire an image after updating the waveforms.

        Can be used in acquisitions where changing waveforms are required,
        but there is additional overhead due to the need to write the
        waveforms into the buffers of the DAQ cards.
        """
        if hasattr(self, "signal_container"):
            self.signal_container.run()

        # Stash current position, channel, timepoint. Do this here, because signal
        # container functions can inject changes to the stage. NOTE: This line is
        # wildly expensive when get_stage_position() does not cache results.
        stage_pos = self.get_stage_position()
        self.data_buffer_positions[self.frame_id][0] = stage_pos.get("x_pos", 0)
        self.data_buffer_positions[self.frame_id][1] = stage_pos.get("y_pos", 0)
        self.data_buffer_positions[self.frame_id][2] = stage_pos.get("z_pos", 0)
        self.data_buffer_positions[self.frame_id][3] = stage_pos.get("theta_pos", 0)
        self.data_buffer_positions[self.frame_id][4] = stage_pos.get("f_pos", 0)

        # Run the acquisition
        try:
            self.active_microscope.turn_on_laser()
            self.active_microscope.daq.run_acquisition()
        except:  # noqa
            self.active_microscope.daq.stop_acquisition()
            if self.active_microscope.current_channel == 0:
                self.stop_acquisition = True
                self.event_queue.put(
                    (
                        "warning",
                        "There is an error happened. Please read the log files for details!",
                    )
                )
                return
            self.active_microscope.daq.prepare_acquisition(
                f"channel_{self.active_microscope.current_channel}"
            )
            self.active_microscope.daq.run_acquisition()
        finally:
            # Ensure the laser is turned off
            self.active_microscope.turn_off_lasers()

        if hasattr(self, "signal_container"):
            self.signal_container.run(wait_response=True)

        self.frame_id = (self.frame_id + 1) % self.number_of_frames

    def run_live_acquisition(self):
        """Stream live image to the GUI.

        Recalculates the waveforms for each image, thereby allowing people to adjust
        acquisition parameters in real-time.
        """
        self.stop_acquisition = False
        while self.stop_acquisition is False and self.stop_send_signal is False:
            self.run_acquisition()
        # Update the stage position.
        # Allows the user to externally move the stage in the continuous mode.
        self.get_stage_position()

    def run_acquisition(self):
        """Run acquisition along with a feature list one time."""
        if not hasattr(self, "signal_container") or self.signal_container.end_flag:
            self.snap_image()
            return

        while (
            not self.signal_container.end_flag
            and not self.stop_send_signal
            and not self.stop_acquisition
        ):
            self.snap_image()
            if not hasattr(self, "signal_container"):
                return
            if self.signal_container.is_closed:
                self.logger.info("Navigate Model - Signal container is closed.")
                self.stop_acquisition = True
                return
        if self.imaging_mode != "live":
            self.stop_acquisition = True

    def change_resolution(self, resolution_value):
        """Switch resolution mode of the microscope.

        Parameters
        ----------
        resolution_value : str
            Resolution mode.
        """
        self.active_microscope.central_focus = None

        former_microscope = self.active_microscope_name
        if resolution_value != self.active_microscope_name:
            self.get_active_microscope()
            self.active_microscope.move_stage_offset(former_microscope)

        # update zoom if possible
        try:
            curr_zoom = self.active_microscope.zoom.zoomvalue
            zoom_value = self.configuration["experiment"]["MicroscopeState"]["zoom"]
            self.active_microscope.zoom.set_zoom(zoom_value)
            self.logger.debug(
                f"Change zoom of {self.active_microscope_name} to {zoom_value}"
            )

            offsets = self.active_microscope.zoom.stage_offsets
            solvent = self.configuration["experiment"]["Saving"]["solvent"]
            if (
                offsets is not None
                and curr_zoom is not None
                and self.active_microscope_name == former_microscope
                and solvent in offsets.keys()
            ):
                # stop stages
                self.active_microscope.stop_stage()
                curr_pos = self.get_stage_position()
                shift_pos = {}
                for axis, mags in offsets[solvent].items():
                    shift_pos[f"{axis}_abs"] = curr_pos[f"{axis}_pos"] + float(
                        mags[curr_zoom][zoom_value]
                    )
                self.move_stage(shift_pos, wait_until_done=True)
            # stop stages and update GUI
            self.stop_stage()

        except ValueError as e:
            self.logger.debug(f"{self.active_microscope_name} - {e}")

        self.active_microscope.ask_stage_for_position = True

    def get_camera_line_interval_and_exposure_time(
        self, exposure_time, number_of_pixel
    ):
        """Get camera line interval time and light sheet exposure time

        Parameters
        ----------
        exposure_time : float
            camera global exposure time
        number_of_pixel: int
            number of pixel in light sheet mode

        Returns
        -------
        exposure_time : float
            Light-sheet mode exposure time (ms).
        camera_line_interval : float
            line interval duration (s).

        """
        return self.active_microscope.camera.calculate_light_sheet_exposure_time(
            exposure_time, number_of_pixel
        )

    def load_images(self, filenames=None):
        """Load/Unload images to the Synthetic Camera

        Parameters
        ----------
        filenames : list
            List of filenames to load.
        """
        self.active_microscope.camera.initialize_image_series(
            self.data_buffer, self.number_of_frames
        )
        self.active_microscope.camera.load_images(filenames)
        self.active_microscope.camera.close_image_series()

    def update_ilastik_setting(
        self, display_segmentation=False, mark_position=True, target_labels=[1]
    ):
        """Update the ilastik setting.

        Parameters
        ----------
        display_segmentation : bool
            Display segmentation.
        mark_position : bool
            Mark position.
        target_labels : list
            Target labels.
        """
        #: bool: Display segmentation.
        self.display_ilastik_segmentation = display_segmentation
        #: bool: Mark position.
        self.mark_ilastik_position = mark_position
        #: list: Target labels.
        self.ilastik_target_labels = target_labels

    def get_microscope_info(self):
        """Return Microscopes device information.

        Returns
        -------
        microscope_info : dict
            Microscope device information.
        """
        microscope_info = {}
        for microscope_name in self.microscopes:
            microscope_info[microscope_name] = self.microscopes[microscope_name].info
        return microscope_info

    def launch_virtual_microscope(self, microscope_name, microscope_config):
        """Launch a virtual microscope.

        Parameters
        ----------
        microscope_name : str
            Name of microscope.
        microscope_config : dict
            Configuration of microscope.

        Returns
        -------
        data_buffer : list
            List of data buffer.
        """

        # create databuffer
        data_buffer = [
            SharedNDArray(shape=(self.img_height, self.img_width), dtype="uint16")
            for i in range(self.number_of_frames)
        ]

        # create virtual microscope
        from navigate.model.devices import (
            SyntheticDAQ,
            SyntheticCamera,  # noqa: F401
            SyntheticGalvo,
            SyntheticFilterWheel,  # noqa: F401
            SyntheticShutter,  # noqa: F401
            SyntheticRemoteFocus,  # noqa: F401
            SyntheticStage,
            SyntheticZoom,  # noqa: F401
            SyntheticMirror,  # noqa: F401
        )

        microscope = Microscope(
            microscope_name, self.configuration, {}, False, is_virtual=True
        )
        microscope.daq = SyntheticDAQ(self.configuration)
        microscope.laser_wavelength = self.microscopes[microscope_name].laser_wavelength
        microscope.lasers = self.microscopes[microscope_name].lasers
        microscope.camera = self.microscopes[microscope_name].camera

        # TODO: lasers
        temp = {
            "zoom": "SyntheticZoom",
            "filter_wheel": "SyntheticFilterWheel",
            "shutter": "SyntheticShutter",
            "remote_focus_device": "SyntheticRemoteFocus",
            "mirror": "SyntheticMirror",
        }

        for k in microscope_config:
            if k.startswith("stage"):
                axis = k[len("stage_") :]
                if microscope_config[k] == "":
                    microscope.stages[axis] = SyntheticStage(
                        microscope_name, None, self.configuration
                    )
                else:
                    microscope.stages[axis] = self.microscopes[microscope_name].stages[
                        axis
                    ]
            elif k.startswith("galvo"):
                if microscope_config[k] == "":
                    microscope.galvo[k] = SyntheticGalvo(
                        microscope_name, None, self.configuration
                    )
                else:
                    microscope.galvo[k] = self.microscopes[microscope_name].galvo[k]
            else:
                if microscope_config[k] == "":
                    exec(
                        f"microscope.{k} = {temp[k]}('{microscope_name}', None, "
                        f"self.configuration)"
                    )
                else:
                    setattr(
                        microscope, k, getattr(self.microscopes[microscope_name], k)
                    )

        # connect virtual microscope with data_buffer
        microscope.update_data_buffer(
            self.img_width, self.img_height, data_buffer, self.number_of_frames
        )

        # add microscope to self.virtual_microscopes
        self.virtual_microscopes[microscope_name] = microscope
        return data_buffer

    def destroy_virtual_microscope(self, microscope_name):
        """Destroy a virtual microscope.

        Parameters
        ----------
        microscope_name : str
            Name of microscope.
        """
        data_buffer = self.virtual_microscopes[microscope_name].data_buffer
        del self.virtual_microscopes[microscope_name]
        # delete shared_buffer
        for i in range(self.number_of_frames):
            data_buffer[i].shared_memory.close()
            data_buffer[i].shared_memory.unlink()
        del data_buffer

    def terminate(self):
        """Terminate the model."""
        self.active_microscope.terminate()
        for microscope_name in self.virtual_microscopes:
            self.virtual_microscopes[microscope_name].terminate()

    def load_feature_list_from_file(self, filename, features):
        """Append feature list from file

        Parameters
        ----------
        filename: str
            filename of the feature list
        features: list
            list of feature names
        """
        module = load_module_from_file(filename[filename.rindex("/") + 1 :], filename)
        for name in features:
            feature = getattr(module, name)
            self.feature_list.append(feature())

    def load_feature_list_from_str(self, feature_list_str):
        """Append feature list from feature_list_str

        Parameters
        ----------
        feature_list_str: str
            str of a feature list
        """
        self.feature_list.append(convert_str_to_feature_list(feature_list_str))

    def load_feature_records(self):
        """Load installed feature lists from system folder

        Note
        ----
            System folcer can be found at '..../.navigate/feature_lists'
        """
        feature_lists_path = get_navigate_path() + "/feature_lists"
        if not os.path.exists(feature_lists_path):
            os.makedirs(feature_lists_path)
            return
        # get __sequence.yml
        if not os.path.exists(f"{feature_lists_path}/__sequence.yml"):
            feature_records = []
        else:
            feature_records = load_yaml_file(f"{feature_lists_path}/__sequence.yml")

        # add non added feature lists
        feature_list_files = [
            temp
            for temp in os.listdir(feature_lists_path)
            if (temp.endswith(".yml") or temp.endswith(".yaml"))
            and os.path.isfile(os.path.join(feature_lists_path, temp))
        ]
        for item in feature_list_files:
            if item == "__sequence.yml":
                continue
            temp = load_yaml_file(f"{feature_lists_path}/{item}")
            add_flag = True
            for feature in feature_records:
                if feature["feature_list_name"] == temp["feature_list_name"]:
                    add_flag = False
                    break
            if add_flag:
                feature_records.append(
                    {
                        "feature_list_name": temp["feature_list_name"],
                        "yaml_file_name": item,
                    }
                )

        i = 0
        while i < len(feature_records):
            temp = feature_records[i]
            if not os.path.exists(f"{feature_lists_path}/{temp['yaml_file_name']}"):
                del feature_records[i]
                continue
            item = load_yaml_file(f"{feature_lists_path}/{temp['yaml_file_name']}")

            if item["module_name"]:
                try:
                    module = load_module_from_file(
                        item["module_name"], item["filename"]
                    )
                    feature = getattr(module, item["module_name"])
                except FileNotFoundError:
                    del feature_records[i]
                    continue
                self.feature_list.append(feature())
            elif item["feature_list"]:
                feature = convert_str_to_feature_list(item["feature_list"])
                self.feature_list.append(feature)
            else:
                del feature_records[i]
                continue
            i += 1
        save_yaml_file(feature_lists_path, feature_records, "__sequence.yml")

    def get_feature_list(self, idx):
        """Get feature list str by index

        Parameters
        ----------
        idx: int
            index of feature list

        Returns
        -------
        feature_list_str: str
            "" if not exist
            string of the feature list
        """
        if idx > 0 and idx <= len(self.feature_list):
            return convert_feature_list_to_str(self.feature_list[idx - 1])
        return ""

    def mark_saving_flags(self, frame_ids):
        """Mark saving flags for the ImageWriter

        Parameters
        ----------
        frame_ids: array
            a list of frame ids
        """
        if not self.data_buffer_saving_flags:
            return
        for id in frame_ids:
            self.data_buffer_saving_flags[id] = True
